import json
import os

from orator import Model
from orator.exceptions.query import QueryException
from orator.orm import belongs_to, scope

from machinable.utils.formatting import str_to_time

try:
    import cPickle as pickle
except ImportError:
    import pickle


class Observation(Model):

    __guarded__ = ["id"]

    __dates__ = ["started", "heartbeat", "finished"]

    @belongs_to
    def task(self):
        from .task import Task

        return Task

    @scope
    def belongs_to_task(self, query, task_id):
        return query.where("task_id", task_id)

    def storage(self):
        return self.task.storage

    def filesystem(self, path=""):
        return self.task.filesystem().opendir(os.path.join(self.path, path))

    def load_file(self, filename, meta=False, filesystem=None, default=None):
        if filesystem is None:
            filesystem = self.filesystem()

        name, ext = os.path.splitext(filename)

        path = "store/" if not meta else ""
        filepath = path + filename

        if not filesystem.exists(filepath):
            return FileNotFoundError if default is None else default

        if ext == ".p":
            with filesystem.open(filepath, "rb") as f:
                data = pickle.load(f)
        elif ext == ".json":
            with filesystem.open(filepath, "r") as f:
                data = json.load(f)

        elif ext == ".npy":
            import numpy as np

            with filesystem.open(filepath, "rb") as f:
                data = np.load(f, allow_pickle=True)

        elif ext == ".txt":
            with filesystem.open(filepath, "r") as f:
                data = f.read()
        else:
            raise ValueError("Invalid file format")

        return data

    @classmethod
    def createFromFile(cls, filesystem, path, **attributes):
        """
        Save a new components read from filesystem an return the instance.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Children
        """
        task_fs = filesystem.opendir(path)

        # attributes

        attributes.setdefault("node", "")
        attributes.setdefault("components", "")
        attributes.setdefault("execution_index", None)

        # status
        if task_fs.isfile("status.json"):
            with task_fs.open("status.json", "r") as f:
                try:
                    status = json.load(f)
                except json.decoder.JSONDecodeError:
                    status = {}
            attributes.setdefault("started", str_to_time(status.get("started")))
            attributes.setdefault("heartbeat", str_to_time(status.get("heartbeat")))
            attributes.setdefault("finished", str_to_time(status.get("finished")))
        try:
            model = cls.first_or_create(path=path, **attributes)
        except QueryException:
            model = None

        return model



import pytest

from machinable.utils.dicts import read_path_dict


def test_read_path_dict():
    t = {
        "one": {"nasty": 1},
        "nested": {"nested": {"nested": 2, "deep": [{"down": 5}, {"rabbit": 3}]}},
        "test": ["case"],
        "foo.bar": 6,
    }

    # dict nesting
    assert read_path_dict(t, "one.nasty") == 1
    assert read_path_dict(t, "nested.nested.nested") == 2
    assert read_path_dict(t, "test") == ["case"]

    with pytest.raises(KeyError):
        read_path_dict(t, "not_existing")

    # array syntax
    assert read_path_dict(t, "test[0]") == "case"
    assert read_path_dict(t, "one[nasty]") == 1
    assert read_path_dict(t, "nested[nested].nested") == 2

    assert read_path_dict(t, "nested[nested].deep[1].rabbit") == 3
    assert read_path_dict(t, "nested[nested][deep][1].rabbit") == 3
    assert read_path_dict(t, "nested[nested][deep][0][down]") == 5

    # escape .
    # todo: fix escaped version
    # assert read_path_dict(t, "foo\.bar") == 6



from .. import execute as _execute
from ..core import Component as _V2Component
from ..core import Mixin
from ..experiment import Experiment as _Experiment
from ..experiment import ExperimentComponent as C
from .observations import Observations


def execute(task, storage=None, seed=None, driver=None):
    return _execute(experiment=task, storage=storage, seed=seed, engine=driver)


class Component(_V2Component):
    @property
    def observer(self):
        return self.store

    @observer.setter
    def observer(self, value):
        self.store = value

    @property
    def children(self):
        return self.components

    @children.setter
    def children(self, value):
        self.components = value


class Task(_Experiment):
    def component(self, node, components=None, resources=None):
        return self.components(node, components, resources)



import copy
import json
from collections import OrderedDict
from typing import Tuple, Type, Union

from ..utils.traits import Jsonable


class ExperimentComponent(Jsonable):
    def __init__(self, name, version=None, checkpoint=None, flags=None):
        """Experiment components

        # Arguments
        name: String, the name of the components as defined in the machinable.yaml
        version: dict|String, a configuration update to override its default config
        checkpoint: String, optional URL to a checkpoint file from which the components will be restored
        flags: dict, optional flags dictionary

        # Examples
        ```python
        import machinable as ml
        component = ml.C('models.linear_regression', {'alpha': 0.1})
        ```
        """
        self.name = name
        self.version = version
        self.checkpoint = checkpoint
        if flags is None:
            flags = {}
        flags = flags.copy()
        if isinstance(checkpoint, str):
            flags["CHECKPOINT"] = checkpoint
        self.flags = flags

    @classmethod
    def create(cls, args: Union[Type, str, Tuple, "ExperimentComponent"]):
        """Creates a components from arguments

        # Returns
        machinable.Component
        """
        if isinstance(args, cls):
            return args

        if args is None:
            return cls(name=None)

        if isinstance(args, str):
            return cls(name=args)

        if isinstance(args, tuple):
            return cls(*args)

        raise ValueError(f"Invalid arguments: {args}")

    def unpack(self):
        if isinstance(self.version, list):
            return [
                __class__(self.name, v, self.checkpoint, self.flags)
                for v in self.version
            ]

        return self

    def serialize(self):
        return (
            self.name,
            copy.deepcopy(self.version),
            self.checkpoint,
            copy.deepcopy(self.flags),
        )

    @classmethod
    def unserialize(cls, serialized):
        if isinstance(serialized, list):
            serialized = tuple(serialized)
        return cls.create(serialized)

    def __str__(self):
        return f"Component({self.name})"

    def __repr__(self):
        if self.name is None:
            return "machinable.C(None)"

        return f"machinable.C({self.name}, version={self.version}, checkpoint={self.checkpoint}, flags={self.flags})"


class Experiment(Jsonable):
    """Defines an execution schedule for available components. The experiment interface is fluent,
    methods can be chained in arbitrary order.

    # Arguments
    components: Optional String or ``tuple`` components definition to add to the experiment

    # Example
    ```python
    import machinable as ml
    linear_regression = ml.Experiment().components('mnist').repeat(3)
    # or using the shorthand where components arguments are passed into the constructor
    linear_regression = ml.Experiment('mnist').repeat(3)
    ```
    """

    def __init__(self):
        self._cache = None
        self._specs = OrderedDict()

    @classmethod
    def create(cls, args):
        """Creates a experiment instance"""
        if isinstance(args, cls):
            return args

        if args is None:
            return cls()

        if isinstance(args, str):
            return cls().components(args)

        if isinstance(args, tuple):
            return cls().components(*args)

    def _spec(self, field, arguments, multiple=False, **kwargs):
        self._cache = None

        if "self" in arguments:
            del arguments["self"]
        value = dict(
            {"field": field, "arguments": arguments, "multiple": multiple}, **kwargs
        )
        if multiple:
            if field not in self._specs:
                self._specs[field] = []
            value["id"] = len(self._specs[field])
            self._specs[field].append(value)
        else:
            self._specs[field] = value

        return self

    @property
    def specification(self):
        if self._cache is not None:
            return self._cache

        # auto-complete
        spec = self._specs.copy()
        spec.setdefault("components", [])
        spec.setdefault("repeats", [])
        spec.setdefault("version", [])

        self._cache = spec

        return spec

    @specification.setter
    def specification(self, specs):
        self._cache = None
        if not isinstance(specs, OrderedDict):
            specs = OrderedDict(specs)
        self._specs = specs

    def serialize(self):
        def serial(x):
            try:
                return x.serialize()
            except AttributeError:
                return x

        serialized = copy.deepcopy(self.specification)
        for i in range(len(serialized["components"])):
            args = serialized["components"][i]["arguments"]
            args["node"] = serial(args["node"])
            args["components"] = [serial(c) for c in args["components"]]

        return serialized

    @classmethod
    def unserialize(cls, serialized):
        for i in range(len(serialized["components"])):
            args = serialized["components"][i]["arguments"]
            args["node"] = ExperimentComponent.unserialize(args["node"])
            args["components"] = [
                ExperimentComponent.unserialize(c) for c in args["components"]
            ]
        task = cls()
        task.specification = serialized
        return task

    # fluent interface

    def copy(self):
        """Returns a copy of the current experiment object
        """
        task = __class__()
        task.specification = self.specification.copy()
        return task

    def component(
        self, name, version=None, checkpoint=None, flags=None, resources=None
    ):
        """Adds a component to the experiment

        # Arguments
        name: String, the name of the components as defined in the machinable.yaml
        version: dict|String, a configuration update to override its default config
        checkpoint: String, optional URL to a checkpoint file from which the components will be restored
        flags: dict, optional flags to be passed to the component
        resources: dict, specifies the resources that are available to the component

        # Examples
        ```python
        import machinable as ml
        experiment = ml.Experiment().component(name='models.linear_regression', version={'alpha': 0.1})
        ```
        """
        return self.components(
            node=ExperimentComponent(name, version, checkpoint, flags),
            resources=resources,
        )

    def components(self, node, components=None, resources=None):
        """Adds a component with sub-components to the experiment

        # Arguments
        node: machinable.ExperimentComponent specifying the node components
        components: optional list of machinable.Component, supplementary components for this node
        resources: dict, specifies the resources that are available to the component

        # Examples
        ```python
        import machinable as ml
        experiment = ml.Experiment().components(
            node=('mnist', version='~shuffled'),                # main component
            components=['resnet', ('resnext', {'lr': 3e-4})]    # sub-components
            resources={'num_gpus': 2}
        )
        ```
        """
        component = ExperimentComponent.create(node).unpack()

        if isinstance(component, list):
            for comp in component:
                self.components(comp, components, resources)
            return self

        if not isinstance(components, list):
            components = [ExperimentComponent.create(components)]

        for i, comp in enumerate(components):
            unpacked_components = ExperimentComponent.create(comp).unpack()
            if isinstance(unpacked_components, list):
                # cross-product with all other components
                for unpacked_component in unpacked_components:
                    reduced = (
                        [components[before] for before in range(i)]
                        + [unpacked_component]
                        + [components[after] for after in range(i + 1, len(components))]
                    )
                    self.components(component, reduced, resources)
                return self

        return self._spec(
            "components",
            {"node": component, "components": components, "resources": resources},
            multiple=True,
        )

    def version(self, config=None, **kwargs):
        """Applies a configuration update

        Selected components use the default configuration as specified in the machinable.yaml.
        Versioning allows for configuration overrides using dictionaries or defined version patterns.
        The versions here are applied globally to all components in this experiment. If you
        want to apply the version locally, use the version parameter in the local method.

        # Arguments
        config: Configuration update represented as dictionary

        # Examples
        ```python
        import machinable as ml
        experiment = ml.Experiment().components('evolution').components('sgd')
        # use a dictionary to override configuration values
        experiment.version({'data': {'shuffle': True}, 'components': {'lr': 0.01}})
        # or kwargs
        experiment.version(data={'shuffle': True}, components={'lr': 0.01})
        # use a specified version of the machinable.yaml
        experiment.version(data='~shuffled')
        # use a mixin configuration
        experiment.version(data='_mnist_')
        ```
        """
        if config is None:
            config_version = copy.deepcopy(kwargs)
        else:
            if isinstance(config, dict):
                config_version = copy.deepcopy(config)
            else:
                config_version = {"components": copy.deepcopy(config)}

        self._spec("version", config_version, multiple=True)

        return self

    # proliferation

    def repeat(self, k, name="REPEAT", mode="independent"):
        """Repeats the experiment k times

        Schedules the current experiment multiple times and injects the flags REPEAT_NUMBER and REPEAT_SEED to the
        node instances.

        # Arguments
        k: Integer, the number of repetitions
        name: String, flag prefix, e.g. '{name}_SEED' etc. Defaults to REPEAT

        # Examples
        ```python
        import machinable as ml
        # five independent runs of the same node
        ml.Experiment().components('regression').repeat(5)
        ```
        """
        return self._spec("repeats", locals(), multiple=True)

    def split(self, k, name="SPLIT", mode="independent"):
        """Splits the experiment k times

        Schedules the current experiment k times and injects appropriate flags SPLIT_NUMBER and SPLIT_SEED to the
        node instances. Note that machinable does not split the nodes automatically. The user
        has to implement the algorithmic splitting based on the flag information. For example, to implement
        a cross-validation algorithm the node should split the data using the split seed in the flag
        SPLIT_SEED and use the split that is specified in the flag SPLIT_NUMBER.

        # Arguments
        k: Integer, the number of splits
        name: String, flag prefix, e.g. '{name}_SEED' etc. Defaults to SPLIT

        # Examples
        ```python
        import machinable as ml
        # five independent runs of the same node
        ml.Experiment().components('regression').repeat(5)
        ```
        """
        return self._spec("repeats", dict(locals(), split=True), multiple=True)

    def tune(self, *args, **kwargs):
        """Schedules for hyperparameter tuning

        Components need to implement on_execute_iteration event that becomes tune training step. The record writer
        can be used as usual. In particular, record fields may be used in stop conditions.

        ::: warning
        Tuning integration is experimental. Please report any issues that you encounter.
        :::

        # Arguments

        The arguments differ based on the used engine.

        - Ray engine: Uses [Ray tune](https://ray.readthedocs.io/en/latest/tune.html) ([Argument reference](https://ray.readthedocs.io/en/latest/tune/api_docs/execution.html#tune-run))
        """
        return self._spec("tune", locals())



import pytest

from machinable import Experiment
from machinable.config.interface import ConfigInterface
from machinable.experiment.parser import parse_experiment
from machinable.project import Project


def test_config_mixin_handler():
    test_project = Project("test_project")

    config = ConfigInterface(test_project.parse_config())

    t = config.get_component("mixexp", version=None, flags=None)["args"]

    # preserved the config
    assert t["hello"] == "there"

    # extended key update
    assert t["key"]["extension"] == "enabled"

    # mixin
    assert t["key"]["overwritten"] == "cool"

    # mixin inheritance
    assert t["key"]["mixing"] == "is"

    # mixin import
    assert t["imported"] == "hello"


def to_config(project, schedule):
    config = ConfigInterface(project.parse_config(), schedule.specification["version"])
    execution_plan = list(parse_experiment(schedule.specification))
    for job_id, (node, children, resources) in enumerate(execution_plan):
        node_config = config.get(node)
        children_config = config.get(children[0])

        if children_config is None:
            return node_config["args"], None

        return node_config["args"], children_config["args"]


def test_config_versioning():
    test_project = Project("./test_project")

    t = Experiment().components(("thenode", {"alpha": -1}))
    e, m = to_config(test_project, t)
    assert e["alpha"] == -1

    t = Experiment().components(("thenode", ({"a": 1}, {"a": 2, "b": 3})))
    e, m = to_config(test_project, t)
    assert e["a"] == 2
    assert e["b"] == 3

    with pytest.raises(KeyError):
        t = Experiment().components(("thenode", "~non-existent"))
        e, m = to_config(test_project, t)

    t = Experiment().components(("thenode", "~one"), ("thechildren", "~two"))
    e, m = to_config(test_project, t)
    assert e["alpha"] == 1
    assert m["alpha"] == 2

    t = Experiment().components(("thenode", ("~three", "~one", "~two")))
    e, m = to_config(test_project, t)
    assert e["alpha"] == 2
    assert e["beta"]["test"]

    # nested
    t = Experiment().components(("thenode", ("~three", "~nested")))
    e, m = to_config(test_project, t)
    assert e["works"]
    assert e["nested"]
    assert e["alpha"] == 4
    assert e["beta"] == "nested"

    t = Experiment().components(("thenode", ("~two", "~nested")))
    e, m = to_config(test_project, t)
    assert e["alpha"] == 2
    assert e["nested"]

    t = Experiment().components(("thenode", ("~three", "~nested", "~nestednested")))
    e, m = to_config(test_project, t)
    assert e["works"]
    assert e["alpha"] == 5
    assert e["q"] == -1
    assert e["beta"] == "overwritten"
    assert e["added"] == "value"

    # mixins
    t = Experiment().components(("thenode", "_trait_"), ("thechildren", "_extended_"))
    e, m = to_config(test_project, t)
    assert e["alpha"] == 0
    assert e["key"]["very"] == "powerful"
    assert m["alpha"] == 0
    assert m["key"]["mixing"] == "is"



import datetime

from orator import DatabaseManager, Schema

from machinable.utils.formatting import prettydict

from ..collections import RecordsCollection
from ..orm.query_builder import QueryBuilder


class Records(RecordsCollection):
    def __init__(self, models, scope="default"):
        self._models = models
        self._scope = scope
        self._db = self._items = None
        self.indexed_columns = {}
        self.refresh()

    def refresh(self, disable_sql=False):
        """Reloads records from the write

        # Arguments
        disable_sql: Boolean to disable the indexing of records into the in-memory SQL database used by self.query
        """
        self._items = self._models.load_file(f"records/{self._scope}.p", meta=True)

        if (
            isinstance(self._items, FileNotFoundError)
            or self._items is FileNotFoundError
        ):
            if self._scope != "default":
                raise FileNotFoundError(
                    f"Records for scope '{self._scope}' could not be found"
                )
            self._items = []
            return

        # create in-memory sql database
        database = ":memory:"

        config = {"records": {"driver": "sqlite", "database": database}}
        self._db = DatabaseManager(config)
        schema = Schema(self._db).connection("records")
        self.indexed_columns = {}
        # schema
        with schema.create("records") as table:
            table.integer("id").unique()
            try:
                if not disable_sql:
                    # use the first row to guess datatype and add supported types to schema
                    for column, value in self._items[0].items():
                        try:
                            # convert numpy scalars into python equivalents
                            import numpy as np

                            if isinstance(
                                value,
                                (
                                    np.float,
                                    np.float16,
                                    np.float32,
                                    np.float64,
                                    np.int,
                                    np.int16,
                                    np.int32,
                                    np.int64,
                                    np.float,
                                    np.float16,
                                    np.float32,
                                    np.float64,
                                ),
                            ):
                                value = value.item()
                        except (ImportError, AttributeError):
                            pass

                        if isinstance(value, datetime.datetime):
                            table.timestamp(column).nullable()
                            self.indexed_columns[column] = datetime.datetime
                        elif isinstance(value, int):
                            table.integer(column).nullable()
                            self.indexed_columns[column] = int
                        elif isinstance(value, bool):
                            table.boolean(column).nullable()
                            self.indexed_columns[column] = bool
                        elif isinstance(value, float):
                            table.float(column).nullable()
                            self.indexed_columns[column] = float
                        elif isinstance(value, str):
                            table.text(column).nullable()
                            self.indexed_columns[column] = str
            except (TypeError, KeyError):
                pass

        def cast(k, v):
            try:
                if isinstance(v, (datetime.datetime, datetime.date)):
                    return v
                return self.indexed_columns[k](v)
            except (TypeError, ValueError):
                return None

        # insert data
        if isinstance(self._items, list):
            for r, row in enumerate(self._items):
                data = {
                    k: cast(k, v) for k, v in row.items() if k in self.indexed_columns
                }
                data.update({"id": r})
                self._db.table("records").insert(data)

    def info(self, echo=False):
        """Returns information about the records as dictionary

        columns: List of columns
        indexed_columns: List of columns that have been indexed and are available in a [query builder](#query)

        # Arguments
        echo: If True, information will be printed
        """
        try:
            columns = [col for col in self.find(0).keys()]
        except TypeError:
            columns = None

        info = {"columns": columns, "indexed_columns": self.indexed_columns}

        if echo:
            print(prettydict(info))

        return info

    @property
    def query(self):
        """Returns a [query builder](#recordsquerybuilder) to build customized queries"""
        return RecordsQueryBuilder(self._db, self._items)

    def find_all(self):
        """Returns all records as a collection"""
        return self.query.get()

    def find(self, record_id=None):
        """Returns a record by ID

        # Arguments
        record_id: Int or list of Int with record IDs. If None, all records are returned

        Returns None or empty collection if record is not found
        """
        if record_id is None:
            return self.query.get()

        if isinstance(record_id, list):
            return self.find_many(record_id)

        try:
            return self._items[record_id]
        except IndexError:
            return None

    def find_many(self, records):
        """Returns a collections of records with a given list of IDs

        # Arguments
        records: List of Record IDs
        """
        return self.query.where_in("id", records).get()

    def __repr__(self):
        return f"Records ({len(self)})"

    def __str__(self):
        return f"Records ({len(self)})"


class RecordsQueryBuilder(QueryBuilder):
    def __init__(self, db, data):
        self._db = db
        self._data = data
        self._query = db.table("records")

    def first(self):
        """Returns the first query result"""
        model = self._query.first()
        return self._data[model.id]

    def _collection(self, models):
        return RecordsCollection(models.transform(lambda model: self._data[model.id]))

    def get(self):
        """Returns the query result"""
        models = self._query.get()
        return self._collection(models)



import copy

from ..config.parser import ModuleClass
from ..core.component import Component as BaseComponent
from ..utils.traits import Jsonable


def recover_class(element):
    if "class" not in element:
        element["class"] = ModuleClass(
            module_name=element["module"], baseclass=BaseComponent
        ).load(instantiate=False)


class Schedule(Jsonable):
    def __init__(self, elements=None):
        self._elements = elements or []
        self._result = []

    def add(
        self, execution_type, component, components, resources, args=None, kwargs=None,
    ):
        self._elements.append(
            [execution_type, component, components, resources, args, kwargs]
        )

    def add_execute(self, component, components, resources, args=None, kwargs=None):
        return self.add("execute", component, components, resources, args, kwargs)

    def add_tune(self, component, components, resources, args=None, kwargs=None):
        return self.add("tune", component, components, resources, args, kwargs)

    def set_result(self, result, index=None):
        if index is None:
            self._result.append(result)
            return

        while len(self._result) <= index:
            self._result.append(None)

        self._result[index] = result

    def serialize(self):
        elements = copy.deepcopy(self._elements)
        serialized = []
        for execution_type, component, components, resources, args, kwargs in elements:
            component.pop("class", None)
            for i in range(len(components)):
                components[i].pop("class", None)
            serialized.append(
                {
                    "execution_type": execution_type,
                    "component": component,
                    "components": components,
                    "resources": resources,
                    "args": args,
                    "kwargs": kwargs,
                }
            )

        return serialized

    @classmethod
    def unserialize(cls, serialized):
        schedule = cls()
        for element in serialized:
            recover_class(element["component"])
            for i in range(len(element["components"])):
                recover_class(element["components"][i])
            schedule.add(**element)
        return schedule

    def iterate(self, storage=None):
        for (
            execution_type,
            component,
            components,
            resources,
            args,
            kwargs,
        ) in self._elements:
            if storage is not None:
                storage = copy.deepcopy(storage)
                storage["uid"] = component["flags"]["UID"]
                yield (
                    execution_type,
                    component,
                    components,
                    storage,
                    resources,
                    args,
                    kwargs,
                )
            else:
                yield (
                    execution_type,
                    component,
                    components,
                    resources,
                    args,
                    kwargs,
                )

    def __getitem__(self, item):
        return self._elements[item]

    def __iter__(self):
        yield from self._elements

    def __len__(self):
        return len(self._elements)



import os
import shutil

import pytest

import machinable.v1 as ml
from machinable.v1.history import get_history

from .generator import generate_data

observations_directory = None


def _setup(debug=False):
    global observations_directory

    # generate data if not available
    if observations_directory is None:
        observations_directory = generate_data(debug=debug)

    mlo = ml.Observations()
    mlo.add(observations_directory)
    return mlo


def _reset_settings():
    config_directory = os.path.expanduser("~/.machinable")
    shutil.rmtree(config_directory, ignore_errors=True)


def test_observations():
    mlo = _setup(debug=False)

    # storages
    mlo.reset()
    assert len(mlo.storages) == 0
    mlo.add(observations_directory)
    assert len(mlo.storages) == 1
    # re-add does not re-add
    for k in range(3):
        mlo.add(observations_directory)
    assert len(mlo.storages) == 1
    assert mlo.remove("never added") is None
    assert len(mlo.storages) == 1
    mlo.remove(observations_directory)
    assert len(mlo.storages) == 0
    mlo.add(observations_directory)

    # standard queries
    assert mlo.find("4NrOUdnAs6A5").config.test
    assert len(mlo.find_by_task("tttttt")) == 4
    assert len(mlo.find_by_task(["tttttt"])) == 4
    assert len(mlo.find_by_node_component("nodes.observations")) == 0
    # len(mlo.find_all())
    assert len(mlo.find_by_most_recent_task()) == 4
    assert len(mlo.find_by_execution(mlo.find("tttttt").first().execution_id)) == 4
    assert len(mlo.find_by_execution(mlo.find("tttttt"))) == 4

    # query builder
    assert len(mlo.query.where_task("tttttt").get()) == 4

    # collections
    assert (
        mlo.find_by_task("tttttt").where("config.to_test", "observations").count() == 4
    )


def test_observation_view():
    mlo = _setup(debug=False)
    observation = mlo.query.where_task("tttttt").first()
    assert observation.storage.endswith("/test_data")
    assert observation.execution_id == observation.task.execution_id
    assert observation.task.id == "tttttt"
    assert observation.task.code_version.project.path is None
    assert observation.flags.NAME == "nodes.observations"
    assert observation.config.to_test == "observations"
    assert len(observation.components) == 0
    assert observation.store("data.json")["observation_id"] > 0
    assert observation.store("test") == 2
    assert observation.store("key") == "value"
    assert "test" in observation.store()
    assert len(observation.store()["$files"])
    assert len(observation.host) == 6
    assert len(observation.get_records_writer()) == 2
    # aliases
    o = mlo.find_by_task_name("second").first()
    # assert o.custom_child_attribute.config.alpha == 0
    # assert o.custom_child_attribute.components == "thechildren"


def test_records_view():
    mlo = _setup(debug=False)
    obs = mlo.query.where_task("tttttt").first()
    records = obs.records
    assert len(records.query.where("constant", ">=", 40).get()) == 6
    # custom records scope
    custom = obs.get_records_writer("validation")
    assert custom.sum("iteration") == 15
    assert records.as_dataframe().size > 0


def test_collections():
    mlo = _setup(debug=False)
    task = mlo.query.where_task("tttttt").get()
    import numpy as np

    def o(x):
        return x.records.pluck("number")

    assert max(task.section(o, reduce=np.var)) > 0
    df = mlo.find_by_task("tttttt").as_dataframe()
    assert df.size == 4 * 12
    print(df.dtypes)
    obs = mlo.query.where_task("tttttt").first()
    num_elements = len(obs.records.pluck("number"))
    with pytest.raises(KeyError):
        obs.records.pluck("not_existing")
    nones = obs.records.pluck_or_none("not_existing")
    assert all([e is None for e in nones])


def test_history():
    h = get_history(file="_test_history")
    h.reset()

    h.add("test")
    assert os.path.isfile(h.url)
    assert len(h) == 1
    h.add("test")
    assert len(h) == 1
    assert len(h.available()) == 0
    h.add("")
    h.add(" ")
    h.add("  ")
    assert len(h) == 1
    h.reset()
    # horizon
    for i in range(200):
        h.add(str(i))
    assert len(h) == 15
    assert sum(map(int, h.data)) == 2880

    h2 = get_history(reload=True, file="_test_history")
    assert len(h2) == 15
    h2.add("")
    assert len(h2) == 15



import os
import shutil

import machinable as ml
from machinable.project import Project, Registration


def test_project_config():
    test_project = Project("./test_project")

    assert test_project.config_filepath == "./test_project/machinable.yaml"

    config = test_project.get_config()

    # default values
    assert "+" in config
    assert "_evaluate" in config

    # import prefix
    import_project = Project("test_project/vendor/fooba")
    assert import_project.import_prefix == "test_project.vendor.fooba"
    import_project = Project("test_project/vendor/fooba/vendor/bedrock")
    assert import_project.import_prefix == "test_project.vendor.fooba.vendor.bedrock"


def test_project_parse_imports():
    test_project = Project("./test_project")

    # remove cache
    cache_path = os.path.join(test_project.directory_path, "vendor", ".cache")
    if os.path.exists(cache_path) and os.path.isdir(cache_path):
        shutil.rmtree(cache_path)

    imports = test_project.parse_imports()

    # import with prefix
    assert "fooba.models.baseline" in imports["components"]
    assert "fooba.experiments.start" in imports["components"]

    # args loaded?
    assert (
        imports["components"]["fooba.models.baseline"]["args"]["overwrite"] == "orginal"
    )
    assert imports["components"]["fooba.experiments.start"]["args"]["test"] == 123

    # args loaded with config inheritance?
    t = imports["components"]["fooba.models.config_inheritance"]["args"]
    assert t["hello"] == "inheritance"
    assert t["overwrite"] == "extended"
    assert t["referenced"] == "extended"

    # args loaded with config dependency inheritance?
    t = imports["components"]["fooba.models.config_dependency_inheritance"]["args"]
    assert t["here"] == "we_go"
    assert t["from_deep_down"] == "huzzah"
    assert t["deep_overwrite"] == "successful"

    # args loaded with chained dependency inheritance?
    t = imports["components"]["fooba.models.chained_inheritance"]["args"]
    assert t["here"] == "we_go"
    assert t["from_deep_down"] == "huzzah"
    assert t["deep_overwrite"] == "early"

    # for experiments:
    t = imports["components"]["fooba.experiments.dependent"]["args"]
    assert t["elephant"] == "at_home"
    assert t["nested"]["elephants"] == "outside"
    assert t["overwrite_me"]["keep"] == "me"
    assert t["overwrite_me"]["nested"] == "level"

    # for mixins
    t = imports["mixins"]["fooba.test"]["args"]
    assert t["imported"] == "hello"
    t = imports["mixins"]["fooba.mixins.nested"]["args"]
    assert t["level"] == -1

    # is importable?
    for k, v in imports["components"].items():
        if k.find("@") != -1:
            continue
        v["class"].load(instantiate=False)


def test_parse_config():
    test_project = Project("./test_project")

    config = test_project.parse_config()

    # args loaded?
    assert config["components"]["childexp"]["args"]["a"] == 2
    assert config["components"]["inherit"]["args"]["blub"] == "bla"
    assert (
        config["components"]["test_dependent"]["args"]["overwrite_me"]["nested"]
        == "here"
    )
    assert config["components"]["cool"]["args"]["bla"] == "blub"

    # args from imports correct?
    assert config["components"]["+.fooba.models.baseline"]["args"]["overwrite"]

    # is importable?
    for k, v in config["components"].items():
        if k.find("@") != -1:
            continue
        v["class"].load(instantiate=False)


def test_project_registration():
    test_project = Project("./test_project")

    assert test_project.has_registration()
    assert isinstance(test_project.registration, Registration)


def test_project_code_backup(helpers):
    helpers.tmp_directory("code_backup")
    target_file = "./_test_data/code_backup/code.zip"
    if os.path.isfile(target_file):
        os.remove(target_file)
    project = Project("./test_project")
    project.backup_source_code(target_file)
    assert os.path.isfile(target_file)


def test_project_get_code_version():
    project = Project("./test_project")
    assert isinstance(project.get_code_version(), dict)



import copy
import datetime
import inspect
import os
import re
from collections import OrderedDict
from inspect import getattr_static
from typing import Dict, List, Optional, Union

from machinable.utils.host import get_host_info

from ..config.mapping import ConfigMap, ConfigMethod, config_map
from ..config.parser import ModuleClass, parse_mixins
from ..store import Store
from ..store.log import Log
from ..store.record import Record
from ..utils.dicts import update_dict
from ..utils.formatting import exception_to_str
from ..utils.traits import Jsonable
from ..utils.utils import apply_seed
from .exceptions import ExecutionException


def set_alias(obj, alias, value):
    if isinstance(alias, str) and not hasattr(obj, alias):
        setattr(obj, alias, value)


def inject_components(component, components, on_create):
    signature = inspect.signature(on_create)

    if components is None:
        components = []

    # default: if there are no given parameters, create each component and assign suggested attribute_
    if len(signature.parameters) == 0:
        for index, c in enumerate(components):
            if not c.component_state.created:
                c.create()
            set_alias(component, c.attribute_, c)
            component.flags.COMPONENTS_ALIAS[c.attribute_] = index

        if component.node is not None:
            set_alias(component, component.node.attribute_, component.node)

        on_create()

        return True

    # user defined
    payload = OrderedDict()
    for index, (key, parameter) in enumerate(signature.parameters.items()):
        if parameter.kind is not parameter.POSITIONAL_OR_KEYWORD:
            # ignore *args and **kwargs
            raise TypeError(
                f"on_create only allows simple positional or keyword arguments"
            )

        if index == 0 and key != "node":
            raise TypeError(
                f"First argument of on_create has to be 'node', received '{key}' instead."
            )

        if key == "node":
            payload["node"] = component.node
        else:
            try:
                c = components[index - 1]
                alias = parameter.name
                if not parameter.name.startswith("_"):
                    c.create()
                    alias = parameter.name[1:]
                payload[parameter.name] = c
                component.flags.COMPONENTS_ALIAS[alias] = index - 1
            except IndexError:
                if parameter.default is parameter.empty:
                    raise TypeError(
                        f"Component {component.__class__.__name__}.on_create requires a components "
                        f"'{parameter.name}' but only {len(components)} were provided."
                    )

                payload[parameter.name] = parameter.default

    on_create(**payload)

    return True


def bind_config_methods(obj, config):
    if isinstance(config, list):
        return [bind_config_methods(obj, v) for v in config]

    if isinstance(config, tuple):
        return (bind_config_methods(obj, v) for v in config)

    if isinstance(config, dict):
        for k, v in config.items():
            if k == "_mixins_":
                continue
            config[k] = bind_config_methods(obj, v)
        return config

    if isinstance(config, str):
        # config method
        fn_match = re.match(r"(?P<method>\w+)\s?\((?P<args>.*)\)", config)
        if fn_match is not None:
            definition = config
            method = "config_" + fn_match.groupdict()["method"]
            args = fn_match.groupdict()["args"]
            if getattr(obj, method, False) is False:
                msg = "Config method %s specified but %s.%s() does not exist." % (
                    definition,
                    type(obj).__name__,
                    method,
                )
                raise AttributeError(msg)
            else:
                return ConfigMethod(obj, method, args, definition)

    return config


class ComponentState(Jsonable):
    def __init__(
        self, created=False, executed=False, destroyed=False, checkpoints=None
    ):
        self.checkpoints = checkpoints or []
        self.created = False
        self.executed = False
        self.destroyed = False

    def serialize(self):
        return {
            "created": self.created,
            "executed": self.executed,
            "destroyed": self.destroyed,
            "checkpoints": self.checkpoints,
        }

    @classmethod
    def unserialize(cls, serialized):
        return cls(**serialized)

    @staticmethod
    def save(component):
        if component.node is not None or component.store is None:
            return False
        component.store.write(
            "state.json",
            {
                "component": component.component_state.serialize(),
                "components": [
                    c.component_state.serialize() for c in component.components
                ]
                if component.components is not None
                else None,
            },
            overwrite=True,
            _meta=True,
        )


class MixinInstance:
    def __init__(self, controller, mixin_class, attribute):
        self.config = {
            "controller": controller,
            "class": mixin_class,
            "attribute": attribute,
        }

    def __getattr__(self, item):
        # lazy-load class
        if isinstance(self.config["class"], ModuleClass):
            self.config["class"] = self.config["class"].load(instantiate=False)

        attribute = getattr(self.config["class"], item, None)

        if attribute is None:
            raise AttributeError(
                f"Mixin '{self.config['class'].__name__}' has no method '{item}'"
            )

        if isinstance(attribute, property):
            return attribute.fget(self.config["controller"])

        if not callable(attribute):
            return attribute

        if isinstance(getattr_static(self.config["class"], item), staticmethod):
            return attribute

        # if attribute is non-static method we decorate it to pass in the controller

        def bound_method(*args, **kwargs):
            # bind mixin instance to controller for mixin self reference
            self.config["controller"].__mixin__ = getattr(
                self.config["controller"], self.config["attribute"]
            )
            output = attribute(self.config["controller"], *args, **kwargs)

            return output

        return bound_method


class Mixin:
    """
    Mixin base class. All machinable mixins must inherit from this base class.
    """

    pass


class Component(Mixin):
    """
    Component base class. All machinable components must inherit from this class.

    ::: tip
    All components are Mixins by default. However, if you want to explicitly mark your components for Mixin-only use
    consider inheriting from the Mixin base class ``machinable.Mixin``.
    :::
    """

    attribute_ = None

    def __init__(self, config: dict = None, flags: dict = None, node=None):
        """Constructs the components instance.

        The initialisation and its events are side-effect free, meaning the application state is preserved
        as if no execution would have happened.
        """
        self.on_before_init(config, flags, node)

        self._node: Optional[Component] = node
        self._components: Optional[List[Component]] = None
        self._store: Optional[Store] = None
        self._actor_config = None
        self.__mixin__ = None
        self.component_state = ComponentState()

        if self.on_init(config, flags, node) is False:
            return

        if config is None:
            config = {}

        # we assign the raw config to allow config_methods to access it
        self._config: ConfigMap = config_map(config)

        # resolve config methods
        self._config: ConfigMap = bind_config_methods(self, config)
        self._config: ConfigMap = ConfigMap(
            self.config, _dynamic=False, _evaluate=self.config.get("_evaluate", True)
        )

        self._flags: ConfigMap = config_map(
            update_dict({"TUNING": False, "COMPONENTS_ALIAS": {}}, flags)
        )

        # bind mixins
        for mixin in parse_mixins(self.config.get("_mixins_"), valid_only=True):
            self.bind(mixin.get("origin", mixin["name"]), mixin["attribute"])

        self.on_after_init()

    def bind(self, mixin, attribute):
        """Binds a mixin to the components

        # Arguments
        mixin: Mixin module or class
        attribute: Attribute name, e.g. _my_mixin_
        """
        if isinstance(mixin, str):
            mixin = ModuleClass(mixin.replace("+.", "vendor."), baseclass=Mixin)
            setattr(self, attribute, MixinInstance(self, mixin, attribute))

    @property
    def config(self) -> ConfigMap:
        """Component configuration"""
        return self._config

    @property
    def flags(self) -> ConfigMap:
        """Component flags"""
        return self._flags

    @property
    def node(self) -> Optional["Component"]:
        """Node component or None"""
        return self._node

    @node.setter
    def node(self, value):
        self._node = value

    @property
    def components(self) -> Optional[List["Component"]]:
        """List of sub components or None"""
        return self._components

    @components.setter
    def components(self, value):
        self._components = value

    @property
    def store(self) -> Store:
        if self._store is None and isinstance(self.node, Component):
            # forward to node store if available
            return self.node.store
        return self._store

    @store.setter
    def store(self, value):
        self._store = value

    @property
    def record(self) -> Record:
        """Record writer instance"""
        return self.store.record

    @property
    def log(self) -> Log:
        """Log writer instance"""
        return self.store.log

    def __getattr__(self, name):
        # Mixins and dynamic attributes are set during construction; this helps suppressing IDE warnings
        raise AttributeError(
            f"{self.__class__.__name__} components has not attribute {name}"
        )

    def dispatch(
        self,
        components_config: List[Dict],
        storage_config: dict,
        actor_config=None,
        lifecycle=True,
    ):
        # Prepares and dispatches the components lifecycle and returns its result
        try:
            self._actor_config = actor_config

            if self.node is None and self.on_seeding() is not False:
                self.set_seed()

            if self.on_init_storage(storage_config) is not False:
                store = Store(config=storage_config)
                self.on_after_init_storage(store)
            else:
                store = None

            if self.on_init_components(components_config) is not False:
                components = []
                index = 0
                for component_config in components_config:
                    if self.on_init_components(component_config, index) is not False:
                        components.append(
                            component_config["class"](
                                config=copy.deepcopy(component_config["args"]),
                                flags=copy.deepcopy(component_config["flags"]),
                                node=self,
                            )
                        )
                        self.on_after_init_components(components[-1], index)
                        index += 1

                self.on_after_init_components(components)
            else:
                components = None

            if not lifecycle:
                return components, store

            self.create(components, store)
            status = self.execute()
            self.destroy()
        except (KeyboardInterrupt, SystemExit):
            status = ExecutionException(
                reason="user_interrupt",
                message="The components execution has been interrupted by the user or system.",
            )
        except BaseException as ex:
            status = ExecutionException(
                reason="exception",
                message=f"The following exception occurred: {ex}\n{exception_to_str(ex)}",
            )

        return status

    def create(self, components=None, store=None):
        """Creates the components

        ::: tip
        This method is triggered automatically. However, child components creation can be suppressed in the on_create
        event of the node components. See [on_create](#on_create) for details.
        :::

        Triggers create events of the lifecycle and restores the components from provided checkpoint if any

        # Arguments
        components: Optional list of child components instances that are used by the components
        store: Optional store instance to be used by the components
        """
        self.on_before_create()

        if store is not None:
            self.store = store

        if components is not None:
            self.components = components

        # prepare child components and invoke on_create event
        inject_components(self, self.components, self.on_create)

        checkpoint = self.flags.get("CHECKPOINT", False)
        if checkpoint:
            self.restore_checkpoint(checkpoint)

        self.on_after_create()

        self.component_state.created = True
        self.component_state.save(self)

    def execute(self):
        """Executes the components

        Triggers execution events and writes execution meta-data

        ::: tip
        Execution is triggered automatically for node components only.
        :::
        """
        self.on_before_execute()

        if self.store:
            self.store.statistics["on_execute_start"] = datetime.datetime.now()
            self.store.write("host.json", get_host_info(), _meta=True)
            self.store.write("component.json", self.serialize(), _meta=True)
            self.store.write(
                "components.json",
                [component.serialize() for component in self.components]
                if self.components
                else [],
                _meta=True,
            )
            self.component_state.save(self)

        try:
            status = self.on_execute()
        except (KeyboardInterrupt, StopIteration) as e:
            status = e

        # if on_iterate is not overwritten ...
        if hasattr(self.on_execute_iteration, "_deactivated") or self.flags.get(
            "TUNING", False
        ):
            # on_execute becomes execute event
            pass
        else:
            # otherwise, we enable the iteration paradigm
            iteration = -1
            while True:
                iteration += 1
                self.flags.ITERATION = iteration
                self.on_before_execute_iteration(iteration)

                try:
                    callback = self.on_execute_iteration(iteration)
                    if self.on_after_execute_iteration(iteration) is not False:
                        # trigger records.save() automatically
                        if (
                            self.store
                            and self.store.has_records()
                            and not self.store.record.empty()
                        ):
                            self.store.record["_iteration"] = iteration
                            self.store.record.save()
                except (KeyboardInterrupt, StopIteration):
                    callback = StopIteration

                if callback is StopIteration:
                    break

        self.on_after_execute()

        self.component_state.executed = True
        self.component_state.save(self)

        return status

    def destroy(self):
        """Destroys the components

        Triggers destroy events

        ::: tip
        This method is triggered automatically.
        :::
        """
        self.on_before_destroy()

        if self.components and len(self.components) > 0:
            for component in self.components:
                if (
                    component.component_state.created
                    and not component.component_state.destroyed
                ):
                    component.destroy()

        self.on_destroy()

        self.component_state.destroyed = True
        self.component_state.save(self)

        if self.store:
            self.store.destroy()

        self.on_after_destroy()

    def set_seed(self, seed=None) -> bool:
        """Applies a global random seed

        # Arguments
        seed: Integer, random seed. If None, self.flags.SEED will be used

        To prevent the automatic seeding, you can overwrite
        the on_seeding event and return False
        """
        if seed is None:
            seed = self.flags.get("SEED")

        return apply_seed(seed)

    def save_checkpoint(self, path: str = None, timestep=None) -> Union[bool, str]:
        """Saves components to a checkpoint

        # Arguments
        path: Path where checkpoints should be saved
        timestep: Optional timestep of checkpoint. If None, timestep will count up automatically

        # Returns
        Filepath of the saved checkpoint or False if checkpoint has not been saved
        """
        if timestep is None:
            timestep: int = len(self.component_state.checkpoints)

        if path is None:
            if not self.store:
                raise ValueError("You need to specify a checkpoint path")

            fs_prefix, basepath = self.store.config["url"].split("://")
            if fs_prefix != "osfs":
                # todo: support non-local filesystems via automatic sync
                raise NotImplementedError(
                    "Checkpointing to non-os file systems is currently not supported."
                )
            checkpoint_path = self.store.get_path("checkpoints", create=True)
            path = os.path.join(os.path.expanduser(basepath), checkpoint_path)

        checkpoint = self.on_save(path, timestep)

        self.component_state.checkpoints.append(
            {"timestep": timestep, "path": path, "checkpoint": checkpoint}
        )
        self.component_state.save(self)

        return checkpoint

    def restore_checkpoint(self, filepath=None):
        """Restores a checkpoint

        # Arguments
        filepath: Checkpoint filepath
        """
        if self.store is not None:
            self.store.log.info(f"Restoring checkpoint {filepath}")
        return self.on_restore(filepath)

    def serialize(self):
        return {
            "config": self.config.toDict(evaluate=True, with_hidden=False),
            "flags": self.flags.toDict(evaluate=True),
        }

    # life cycle

    def on_seeding(self):
        """Lifecycle event to implement custom seeding

        Return False to prevent the default seeding procedure
        """
        pass

    def on_before_init(self, config=None, flags=None, node=None):
        """Lifecycle event triggered before components initialisation"""
        pass

    def on_init(self, config=None, flags=None, node=None):
        """Lifecycle event triggered during components initialisation

        Return False to prevent the default configuration and flag instantiation
        """
        pass

    def on_after_init(self):
        """Lifecycle event triggered after components initialisation"""
        pass

    def on_before_create(self):
        """Lifecycle event triggered before components creation"""
        pass

    def on_create(self):
        """Lifecycle event triggered during components creation

        The method can declare arguments to handle components explicitly. For example, the signature
        ``on_create(self, node, alias_of_child_1, alias_of_child_2=None)`` declares that components
        accepts two sub components with alias ``alias_of_child_1`` and ``alias_of_child_2`` where
        the latter is declared optional. If the alias starts with an underscore ``_`` the components lifecycle
        will not be triggered automatically.

        The function signature may also be used to annotate expected types, for example:

        ```python
        on_create(self, node: DistributedExperiment, model: DistributedModel = None):
            self.experiment = node
            self.model = model
        ```
        Note, however, that the type annotation will not be enforced.
        """
        pass

    def on_save(self, checkpoint_path: str, timestep: int) -> Union[bool, str]:
        """Implements the checkpoint functionality of the components

        # Arguments
        checkpoint_path: String, target directory
        timestep: Integer, counting number of checkpoint

        # Returns
        Returns the name of the checkpoint file or path. Return False to indicate that checkpoint has not been saved.
        """
        pass

    def on_restore(self, checkpoint_file: str):
        """Implements the restore checkpoint functionality of the components

        # Arguments
        checkpoint_path: String, source directory
        timestep: Integer, counting number of checkpoint

        # Returns
        Return False to indicate that checkpoint has not been restored.
        """
        pass

    def on_init_components(self, components_config, index: Optional[int] = None):
        """Lifecycle event triggered during components initialisation

        Return False to prevent the components instantiation"""
        pass

    def on_after_init_components(self, components, index: Optional[int] = None):
        """Lifecycle event triggered after components initialisation"""
        pass

    def on_init_storage(self, storage_config: Dict):
        """Lifecycle event triggered at write initialisation

        Return False to prevent the default write instantiation
        """
        pass

    def on_after_init_storage(self, storage: Store):
        """Lifecycle event triggered after store initialisation"""
        pass

    def on_after_create(self):
        """Lifecycle event triggered after components creation"""
        pass

    def on_before_execute(self):
        """Lifecycle event triggered before components execution"""
        pass

    def on_execute(self):
        """Lifecycle event triggered at components execution

        ::: tip
        This event and all other execution events are triggered in node components only
        :::
        """
        pass

    def on_before_execute_iteration(self, iteration: int):
        """Lifecycle event triggered before an execution iteration
        """
        pass

    def on_execute_iteration(self, iteration: int):
        """Lifecycle event triggered at execution iteration

        Allows to implement repeating iterations

        The method is repeatedly called until StopIteration is returned or raised; iteration is increased on each call
        The iteration number is also available class-wide under self.flags.ITERATION

        # Arguments
        iteration: Integer, iteration number that automatically increases in every iteration starting from 0
        """
        pass

    # disable unless overridden
    on_execute_iteration._deactivated = True

    def on_after_execute_iteration(self, iteration: int):
        """Lifecycle event triggered after execution iteration"""
        pass

    def on_after_execute(self):
        """Lifecycle event triggered after execution"""
        pass

    def on_before_destroy(self):
        """Lifecycle event triggered before components destruction"""
        pass

    def on_destroy(self):
        """Lifecycle event triggered at components destruction"""
        pass

    def on_after_destroy(self):
        """Lifecycle event triggered after components destruction"""
        pass


class FunctionalComponent:
    def __init__(self, function, node_config=None, node_flags=None):
        self.function = function
        self.node = {"config": None, "flags": None}
        self.set_config(node_config, node_flags)
        self._argument_structure = OrderedDict()

    def __call__(self, config=None, flags=None):
        # virtual constructor
        self.set_config(config, flags)
        return self

    def set_config(self, config, flags):
        if config is None:
            config = {}
        if flags is None:
            flags = {}
        self.node["config"] = copy.deepcopy(config)
        self.node["flags"] = copy.deepcopy(flags)

    def dispatch(self, components_config, storage_config, actor_config=None):
        apply_seed(self.node["flags"].get("SEED"))
        self.node["flags"]["ACTOR"] = actor_config
        components = [
            {"config": copy.deepcopy(c["args"]), "flags": copy.deepcopy(c["flags"]),}
            for c in components_config
        ]

        # analyse argument structure
        signature = inspect.signature(self.function)
        payload = OrderedDict()

        for index, (key, parameter) in enumerate(signature.parameters.items()):
            if parameter.kind is not parameter.POSITIONAL_OR_KEYWORD:
                # disallow *args and **kwargs
                raise TypeError(
                    f"Component only allows simple positional or keyword arguments,"
                    f"for example my_component(node, components, store)"
                )

            if key == "node":
                payload["node"] = config_map(self.node)
            elif key == "components":
                payload["components"] = [
                    config_map(component) for component in components
                ]
            elif key == "store" or key == "_store":
                if key == "_store":
                    payload["store"] = storage_config
                else:
                    store = Store(storage_config)
                    store.statistics["on_execute_start"] = datetime.datetime.now()
                    store.write("host.json", get_host_info(), _meta=True)
                    store.write(
                        "component",
                        {"config": self.node["config"], "flags": self.node["flags"]},
                        _meta=True,
                    )
                    store.write(
                        "components",
                        [
                            {"config": c["config"], "flags": c["flags"]}
                            for c in components
                        ],
                        _meta=True,
                    )
                    payload["store"] = store
            else:
                raise ValueError(
                    f"Unrecognized argument: '{key}'. "
                    f"Components take the following arguments: "
                    f"'node', 'components' and 'store'"
                )

        try:
            return self.function(**payload)
        except (KeyboardInterrupt, SystemExit):
            status = ExecutionException(
                reason="user_interrupt",
                message="The components execution has been interrupted by the user or system.",
            )
        except BaseException as ex:
            status = ExecutionException(
                reason="exception",
                message=f"The following exception occurred: {ex}\n{exception_to_str(ex)}",
            )

        return status



class QueryBuilder:
    @property
    def object(self):
        """Returns the query object"""
        return self._query

    def _field_mapping(self, field):
        return field

    def where(self, field, operator="=", value=None, boolean="and"):
        """"""
        field = self._field_mapping(field)
        self._query = self._query.where(field, operator, value, boolean)

        return self

    def or_where(self, field, operator="=", value=None):
        """"""
        return self.where(field, operator, value, "or")

    def where_between(self, field, values, boolean="and", negate=False):
        """"""
        field = self._field_mapping(field)
        self._query = self._query.where_between(field, values, boolean, negate)

        return self

    def or_where_between(self, field, values):
        """"""
        return self.where_between(field, values, "or")

    def where_not_between(self, field, values, boolean="and"):
        """"""
        return self.where_between(field, values, boolean, True)

    def or_where_not_between(self, field, values):
        """"""
        return self.where_not_between(field, values, "or")

    def where_in(self, field, values, boolean="and", negate=False):
        """"""
        field = self._field_mapping(field)
        self._query = self._query.where_in(field, values, boolean, negate)

    def or_where_in(self, field, values):
        """"""
        return self.where_in(field, values, "or")

    def where_not_in(self, field, values, boolean="and"):
        """"""
        return self.where_in(field, values, boolean, True)

    def or_where_not_in(self, field, values):
        """"""
        return self.where_not_in(field, values, "or")

    def where_has(self, relation, extra, operator=">=", count=1):
        """"""
        self._query = self._query.where_has(relation, extra, operator, count)

        return self

    def order_by(self, field, direction="asc"):
        """"""
        field = self._field_mapping(field)
        self._query = self._query.order_by(field, direction)
        return self

    def skip(self, value):
        """"""
        self._query = self._query.skip(value)
        return self

    def take(self, value):
        """"""
        self._query = self._query.take(value)
        return self

    def union(self, query, all=False):
        """"""
        self._query.union(query, all)
        return self

    def union_all(self, query):
        """"""
        return self.union(query, True)



import os
import shutil

import pytest


class Helpers:
    @staticmethod
    def tmp_directory(append=""):
        path = os.path.join("./_test_data/", append)
        shutil.rmtree(path, ignore_errors=True)
        os.makedirs(path)
        return path


@pytest.fixture
def helpers():
    return Helpers



import os
import sys

import pytest

import machinable as ml
from machinable import Component, Engine
from machinable.core.component import inject_components


def test_core_config_method():
    ml.execute("configmethods", project="./test_project")


def test_on_iterate():
    class TestIterate(ml.Component):

        ITER = 0
        USE_RECORDS = False

        def on_execute_iteration(self, iteration):
            self.ITER = iteration

            if self.USE_RECORDS:
                self.record["hello"] = iteration
            if iteration == 5:
                return StopIteration

    iterator = TestIterate()
    iterator.dispatch([], {"uid": "12345"})
    iterator.create()
    iterator.execute()

    # iteration working
    assert iterator.ITER == 5

    # there should be no records
    assert "default" not in iterator.store._record_writers

    # repeat with records writing
    iterator = TestIterate()
    iterator.dispatch([], {"uid": "12345"})
    iterator.create()
    iterator.USE_RECORDS = True
    iterator.execute()

    # records should now be available
    assert "default" in iterator.store._record_writers
    # records should be empty because they have been saved
    assert iterator.record.empty()
    # there should be 5 records
    assert len(iterator.record.history) == 6
    # they should contain _iteration
    print(iterator.record.history)
    for i in range(len(iterator.record.history)):
        assert iterator.record.history[i]["_iteration"] == i


def test_interaction():
    ml.execute(
        ml.Experiment().components("thenode", "workers.interactive"),
        project="./test_project",
    )


def test_component_injection():
    class DummyChild(ml.Component):
        def __init__(self, should_be_created=True, attribute=None):
            super().__init__()
            self.attribute_ = attribute
            self.should_be_created = should_be_created

        def create(self):
            assert self.should_be_created

    class TestInject(ml.Component):
        def on_create_no_params(self):
            pass

        def on_create_manual(self, node, first, _second, third):
            assert node is None
            assert isinstance(first, DummyChild)
            assert isinstance(_second, DummyChild)
            assert isinstance(third, DummyChild)

        def on_create_manual_with_node(self, node, first, _second, third):
            assert isinstance(node, DummyChild)
            assert isinstance(first, DummyChild)
            assert isinstance(_second, DummyChild)
            assert isinstance(third, DummyChild)

    # components are created automatically with suggested attribute_
    t = TestInject()
    assert inject_components(
        t, [DummyChild(), DummyChild(attribute="model")], t.on_create_no_params
    )
    assert isinstance(t.model, DummyChild)

    # components are created based on signature
    t = TestInject(node=DummyChild(attribute="experiment"))
    assert inject_components(
        t, [DummyChild(), DummyChild(False), DummyChild()], t.on_create_manual_with_node
    )


def test_exception_handling():
    sys.path.insert(0, os.path.join(os.getcwd(), "test_project"))
    from test_project.failure.exceptions import ExceptionsComponent

    ex = ExceptionsComponent()
    status = ex.dispatch([], {"uid": "12345"})
    assert isinstance(status, ml.core.exceptions.ExecutionException)

    ml.execute(
        ml.Experiment().components("failure.exceptions"), project="./test_project"
    )

    # a failure does not crash others
    import ray

    ray.init(ignore_reinit_error=True)
    ml.execute(
        ml.Experiment()
        .components("failure.exceptions")
        .components("thenode")
        .repeat(2),
        project="./test_project",
    )


def test_mixins():
    # set test project path
    sys.path.insert(0, os.path.join(os.getcwd(), "test_project"))

    # config only mixin
    component = Component({"_mixins_": ["extended", "+.fooba.test"]}, {})
    assert getattr(component, "extended", True)
    assert getattr(component, "fooba_test", True)

    # module mixins

    component = Component(
        {"_mixins_": ["mixin_module", "+.fooba.mixins.nested"]}, {"BOUND": "component"}
    )
    assert getattr(component, "_mixin_module_", None) is not None
    assert (
        component._mixin_module_.is_bound("correctly") == "bound_to_component_correctly"
    )
    with pytest.raises(AttributeError):
        component._mixin_module_.non_existent("call")

    assert getattr(component, "_fooba_mixins_nested_", None) is not None
    assert component._fooba_mixins_nested_.hello() == "component"
    with pytest.raises(AttributeError):
        component._fooba_mixins_nested_.non_existent("call")

    assert component._mixin_module_.key_propery == 1

    # de-alias via origin
    config = {
        "_mixins_": [
            "mixin_module",
            "+.fooba.mixins.nested",
            {"name": "+.fooba.nested", "origin": "+.fooba.mixins.nested"},
        ]
    }
    component = Component(config, {"BOUND": "component"})
    assert (
        component._mixin_module_.is_bound("correctly") == "bound_to_component_correctly"
    )

    # this-referencing
    assert (
        component._mixin_module_.this_reference("correctly")
        == "bound_to_component_and_referenced_correctly"
    )
    assert component._mixin_module_.this_attribute() == "works"
    assert component._mixin_module_.this_static("works") == "works"


def test_hidden_mixins():
    sys.path.insert(0, os.path.join(os.getcwd(), "test_project"))

    # hidden mixins that are only part of the imported project but not referenced in the project that imports them
    ml.execute(ml.Experiment().components("inherited_mixin"), project="./test_project")
    ml.execute(
        ml.Experiment().components("direct_mixin_inheritance"), project="./test_project"
    )



from machinable import Component


class MyComponent(Component):
    def on_create(self):
        print(self.config.config_value)
        print(self.config.nested.value)
        print(self.config["nested"]["value"])



class BaseView:
    def __init__(self, model, cache=None):
        self._model = model
        # shared cache across interfaces with lookup by their sqlite database id
        if cache is None:
            cache = {}
        self._cache = cache

    def as_json(self):
        """Returns a json representation of the data"""
        return self._model.to_json()

    def serialize(self):
        """Serializes the data"""
        return self._model.serialize()

    def _lazyload(self, field, loader):
        if self._model.id not in self._cache:
            self._cache[self._model.id] = {}

        if field not in self._cache[self._model.id]:
            self._cache[self._model.id][field] = loader(self._model)

        return self._cache[self._model.id][field]



# This file contains modified 3rd party source code from
# https://github.com/drgrib/dotmap
# The copyright and license agreement can be found in the ThirdPartyNotices.txt file at the root of this repository.

from collections import OrderedDict
from inspect import ismethod
from json import dumps
from pprint import pprint

from dotmap import DotMap


def config_map(d=None):
    if d is None:
        d = {}
    return ConfigMap(d, _dynamic=False, _evaluate=False, _evaluated=True)


class ConfigMethod(object):
    def __init__(self, obj, method, args, definition):
        self._ = {"obj": obj, "method": method, "args": args, "definition": definition}

    def evaluate(self):
        # disable config evaluation
        state = self._["obj"].config._evaluate
        self._["obj"].config._evaluate = False
        try:
            costate = self._["obj"].co.config._evaluate
            self._["obj"].co.config._evaluated = False
        except AttributeError:
            costate = None

        if self._["args"] == "":
            value = getattr(self._["obj"], self._["method"])()
        else:
            # Yes, using eval is evil but in this case I suppose there is not enough at stake to justify
            # the implementation of a proper parser
            obj = self._["obj"]  # noqa: F841
            value = eval("obj." + self._["method"] + "(" + self._["args"] + ")")

        # reset config evaluation
        self._["obj"].config._evaluate = state
        if costate is not None:
            self._["obj"].co.config._evaluate = costate

        if isinstance(value, dict):
            return ConfigMap(value, _dynamic=False)

        return value

    def __repr__(self):
        return f"ConfigMethod(method={self._['obj'].__class__.__name__}.{self._['method']}, args='{self._['args']}')"

    def __str__(self):
        return self.__repr__()

    def __call__(self, *args, **kwargs):
        return self.evaluate()


class ConfigMap(DotMap):
    def __init__(self, *args, **kwargs):
        self._map = OrderedDict()
        self._dynamic = True
        self._evaluate = True
        self._evaluated = False
        if kwargs:
            if "_dynamic" in kwargs:
                self._dynamic = kwargs["_dynamic"]
            if "_evaluate" in kwargs:
                self._evaluate = kwargs["_evaluate"]
            if "_evaluated" in kwargs:
                self._evaluated = kwargs["_evaluated"]
        if args:
            d = args[0]
            # for recursive assignment handling
            trackedIDs = {id(d): self}
            if isinstance(d, dict):
                for k, v in self.__call_items(d):
                    if isinstance(v, dict):
                        if id(v) in trackedIDs:
                            v = trackedIDs[id(v)]
                        else:
                            v = self.__class__(
                                v,
                                _dynamic=self._dynamic,
                                _evaluate=self._evaluate,
                                _evaluated=self._evaluated,
                            )
                            trackedIDs[id(v)] = v
                    if type(v) is list:
                        lst = []
                        for i in v:
                            n = i
                            if isinstance(i, dict):
                                n = self.__class__(
                                    i,
                                    _dynamic=self._dynamic,
                                    _evaluate=self._evaluate,
                                    _evaluated=self._evaluated,
                                )
                            lst.append(n)
                        v = lst
                    self._map[k] = v
        if kwargs:
            for k, v in self.__call_items(kwargs):
                if k not in ("_dynamic", "_evaluate", "_evaluated"):
                    self._map[k] = v

    def toDict(self, evaluate=None, with_hidden=True):
        if evaluate is None:
            evaluate = bool(self._evaluate)
        d = {}
        for k, v in self.items():
            if (
                with_hidden is False
                and isinstance(k, str)
                and (k.startswith("_") and not k.endswith("_"))
            ):
                continue
            if evaluate and isinstance(v, ConfigMethod):
                v = v.evaluate()
            if issubclass(type(v), DotMap):
                # bizarre recursive assignment support
                if id(v) == id(self):
                    v = d
                else:
                    v = v.toDict(evaluate=evaluate, with_hidden=with_hidden)
            elif type(v) in (list, tuple):
                l = []
                for i in v:
                    n = i
                    if issubclass(type(i), DotMap):
                        n = i.toDict(evaluate=evaluate, with_hidden=with_hidden)
                    l.append(n)
                if type(v) is tuple:
                    v = tuple(l)
                else:
                    v = l
            d[k] = v
        return d

    def pprint(self, pformat="json"):
        if pformat == "json":
            print(dumps(self.toDict(), indent=4, sort_keys=True))
        else:
            pprint(self.toDict())

    def __call_items(self, obj):
        if hasattr(obj, "iteritems") and ismethod(getattr(obj, "iteritems")):
            return obj.iteritems()
        else:
            return obj.items()

    def copy(self):
        return self.__class__(
            self,
            _dynamic=self._dynamic,
            _evaluate=self._evaluate,
            _evaluated=self._evaluated,
        )

    def get_versioning(self):
        return [k[1:] for k in self.keys() if k.startswith("~")]

    def __getitem__(self, k, evaluate=None):
        if (
            k not in self._map
            and self._dynamic
            and k != "_ipython_canary_method_should_not_exist_"
        ):
            # automatically extend to new DotMap
            self[k] = self.__class__()

        var = self._map[k]

        # evaluate
        if evaluate is None:
            evaluate = self._evaluate

        if evaluate:
            if isinstance(var, ConfigMethod):
                var = var.evaluate()

        return var

    def __setattr__(self, k, v):
        if k in {
            "_map",
            "_dynamic",
            "_evaluate",
            "_evaluated",
            "_ipython_canary_method_should_not_exist_",
            "get_versioning",
        }:
            super(DotMap, self).__setattr__(k, v)
        else:
            self[k] = v

    def __getattr__(self, k):
        if k in {
            "_map",
            "_dynamic",
            "_evaluate",
            "_evaluated",
            "_ipython_canary_method_should_not_exist_",
            "get_versioning",
        }:
            return super(DotMap, self).__getattr__(k)

        try:
            v = super(self.__class__, self).__getattribute__(k)
            return v
        except AttributeError:
            pass

        return self[k]


_reserved_keys = ["_map", "_dynamic", "_evaluate", "_evaluated"]
_used_keys = [
    "toDict",
    "pprint",
    "items",
    "next",
    "empty",
    "values",
    "parseOther",
    "clear",
    "copy",
    "get",
    "has_key",
    "keys",
    "pop",
    "popitem",
    "setdefault",
    "update",
    "fromkeys",
    "bannerStr",
    "_getSubMapStr",
    "_getSubMapDotList",
    "_getValueStr",
    "_getListStr",
]



import os
import shutil

import machinable as ml
from machinable.engine import Engine


def generate_data(path=None, debug=False):
    # delete and re-create data directory
    if path is None:
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "test_data")

    if debug and os.path.exists(path):
        # do not re-generate if existing
        return path

    shutil.rmtree(path, ignore_errors=True)
    os.makedirs(path)

    # run data generator
    ml.execute(
        ml.Experiment().components(("nodes.observations", {"id": 1})).repeat(3),
        path,
        project="./test_project",
    )
    ml.execute(
        ml.Experiment()
        .components(("nodes.observations", {"id": 2}), "thechildren")
        .repeat(2),
        path,
        project="./test_project",
    )

    ml.execute(
        ml.Experiment()
        .components(("nodes.observations", {"id": 3, "test": True}))
        .repeat(4),
        path,
        seed="tttttt",
        project="./test_project",
    )

    ml.execute(
        ml.Experiment().components(("nodes.observations", {"id": 4, "corrupt": True})),
        path,
        seed="corupt",
        project="./test_project",
    )

    # corrupt some data
    shutil.rmtree(os.path.join(path, "corupt"), ignore_errors=True)

    return path



import os
import shutil

import numpy as np

from machinable import Experiment, execute
from machinable.store import Store


def test_store_writer():
    store = Store({"uid": "12345"})

    # write
    store.write("test.txt", "test me")
    f = os.path.join(store.config["group"], store.config["uid"], "store", "test.txt")
    assert store.filesystem.readtext(f) == "test me"
    store.write("test.npy", np.ones([5]))
    store.write("test.p", np.ones([5]))
    store.write("test.json", [1, 2, 3])
    store.write("dir/test.txt", "subdirectory")
    f = os.path.join(
        store.config["group"], store.config["uid"], "store", "dir/test.txt"
    )
    assert store.filesystem.readtext(f) == "subdirectory"
    f = os.path.join(store.config["group"], store.config["uid"], "store.json")
    store.write("test", True)
    assert store.filesystem.readtext(f) == '{"test": true}'
    store.write("bla", 1)
    assert store.filesystem.readtext(f) == '{"test": true, "bla": 1}'

    # observations
    store.record["test"] = 1
    assert store.record["test"] == 1

    # log
    store.log.info("test")


def test_records_timing():
    execute(Experiment().components("timings"), project="./test_project")


def test_output_redirection(capsys, helpers):
    storage = helpers.tmp_directory("output_redirection")

    for mode in ["SYS_AND_FILE", "FILE_ONLY", "DISCARD"]:
        if os.path.exists(storage):
            shutil.rmtree(storage, ignore_errors=True)

        print("non-captured")
        o = Store({"uid": "654321", "url": storage, "output_redirection": mode})
        print("captured")
        o.destroy()
        print("non-captured-again")
        if mode == "DISCARD":
            assert not os.path.isfile(os.path.join(storage, "654321/output.log"))
        else:
            with open(os.path.join(storage, "654321/output.log"), "r") as f:
                assert f.read() == "captured\n"

        assert (
            capsys.readouterr().out
            == {
                "SYS_AND_FILE": "non-captured\ncaptured\nnon-captured-again\n",
                "FILE_ONLY": "non-captured\nnon-captured-again\n",
                "DISCARD": "non-captured\nnon-captured-again\n",
            }[mode]
        )



import json
import os
import pickle

from fs import open_fs
from fs.opener.parse import parse_fs_url


class Directory:
    def __init__(self, url):
        if "://" not in url:
            url = "osfs://" + url
        self._url = url
        resource = os.path.normpath(parse_fs_url(self._url).resource)
        self._path = os.path.basename(resource)
        self._uid = None
        if len(self._path) == 12:
            # switch to parent
            self._uid = self._path
            self._path = os.path.basename(os.path.dirname(resource))
            self._url = self._url.replace("/" + self._uid, "")
        if len(self._path) != 6:
            raise ValueError("The provided path is not a valid storage directory")

        self.filesystem = open_fs(self._url, create=False)
        # todo: implement automatic caching layer

    def set_default_uid(self, uid):
        self._uid = uid

    def load_file(self, filename, uid=None, default=None):
        if uid is None:
            uid = self._uid or ""

        if not isinstance(uid, str):
            raise ValueError(f"Invalid UID: {uid}")

        name, ext = os.path.splitext(filename)

        filepath = os.path.join(uid, filename)

        if not self.filesystem.exists(filepath):
            return FileNotFoundError if default is None else default

        if ext == ".p":
            with self.filesystem.open(filepath, "rb") as f:
                data = pickle.load(f)
        elif ext == ".json":
            with self.filesystem.open(filepath, "r") as f:
                data = json.load(f)
        elif ext == ".npy":
            import numpy as np

            with self.filesystem.open(filepath, "rb") as f:
                data = np.load(f, allow_pickle=True)
        elif ext in [".txt", ".log"]:
            with self.filesystem.open(filepath, "r") as f:
                data = f.read()
        else:
            raise ValueError(f"Invalid file format: {ext}")

        return data



import time

from machinable import Component


class TheNode(Component):
    def on_create(self):
        if self.config.timeout > 0:
            print("sleeping for a while")
            time.sleep(self.config.timeout)
        print("hello from the node")

    def config_through_config_method(self, arg):
        return arg



import threading
import typing as T

from observable import Observable


class Events(Observable):
    def __init__(self) -> None:
        super().__init__()
        self._heartbeat = None

    def trigger(self, event: str, *args: T.Any, **kw: T.Any) -> list:
        # upstream pending on issue https://github.com/timofurrer/observable/issues/17
        """Triggers all handlers which are subscribed to an event.
        Returns True when there were callbacks to execute, False otherwise."""

        callbacks = list(self._events.get(event, []))
        return [callback(*args, **kw) for callback in callbacks]

    def heartbeats(self, seconds=10):
        if self._heartbeat is not None:
            self._heartbeat.cancel()

        if seconds is None or int(seconds) == 0:
            # disable heartbeats
            return

        def heartbeat():
            t = threading.Timer(seconds, heartbeat)
            t.daemon = True
            t.start()
            self.trigger("heartbeat")
            return t

        self._heartbeat = heartbeat()



import json

from orator import Model
from orator.orm import belongs_to, has_many, scope

from machinable.utils.formatting import str_to_time


class Task(Model):

    __guarded__ = ["id"]

    __dates__ = ["started", "heartbeat", "finished"]

    @belongs_to
    def storage(self):
        from .storage import Storage

        return Storage

    @has_many
    def observations(self):
        from .observation import Observation

        return Observation

    @scope
    def belongs_to_storage(self, query, storage_id):
        return query.where("storage_id", storage_id)

    def filesystem(self):
        return self.storage.filesystem().opendir(self.path)

    def index(self, filesystem=None):
        from .observation import Observation

        if filesystem is None:
            filesystem = self.filesystem()

        # delete all stored observations of this experiment before re-indexing
        Observation.belongs_to_task(self.id).delete()

        for directory in filesystem.listdir(""):
            if len(directory) == 12:
                Observation.createFromFile(filesystem, path=directory, task_id=self.id)

    @classmethod
    def createFromFile(cls, filesystem, path, **attributes):
        """
        Save a new experiment read from filesystem and return the instance.

        :param attributes: The instance attributes
        :type attributes: dict

        :return: The new instance
        :rtype: Task
        """
        task_fs = filesystem.opendir(path)

        if not task_fs.isfile("execution.json"):
            return False

        # attributes
        with task_fs.open("execution.json", "r") as f:
            try:
                data = json.load(f)
            except json.decoder.JSONDecodeError:
                data = {}

        attributes.setdefault("name", data.get("name"))
        attributes.setdefault("task_id", data.get("id"))
        attributes.setdefault("execution_id", data.get("timestamp"))
        attributes.setdefault("seed", data.get("seed"))
        attributes.setdefault("tune", data.get("tune"))
        attributes.setdefault("execution_cardinality", len(data.get("schedule", [])))
        attributes.setdefault("rerun", int(data.get("rerun", 0)))
        attributes.setdefault("code_backup", bool(data.get("code_backup")))
        attributes.setdefault("code_version", json.dumps(data.get("code_version")))

        attributes.setdefault("observations_count", len(data.get("schedule", [])))
        attributes.setdefault("started", str_to_time(data.get("started_at")))

        model = cls.first_or_create(path=path, **attributes)

        model.index(task_fs)

        return model


def deleting(task):
    # clean up observations
    from .observation import Observation

    Observation.belongs_to_task(task.id).delete()


Task.deleting(deleting)



import os
import platform
import socket

import pkg_resources

_getters = {}


def register_host_info(function):
    name = function.__name__
    if name.startswith("_"):
        name = name[1:]
    _getters[name] = function


def get_host_info():
    return {name: getter() for name, getter in _getters.items()}


@register_host_info
def _network_name():
    return platform.node()


@register_host_info
def _hostname():
    return socket.gethostname()


@register_host_info
def _machine():
    return platform.machine()


@register_host_info
def _python_version():
    return platform.python_version()


@register_host_info
def _environ():
    return os.environ.copy()


@register_host_info
def _machinable_version():
    try:
        return pkg_resources.require("machinable")[0].version
    except:  # noqa: E722
        return "unknown"



import numpy as np

import machinable as ml

random_search = (
    ml.Experiment()
    .components("optimization")
    .tune(
        stop={"episodes_total": 50},
        num_samples=50,
        config={
            "stepsize": lambda spec: np.random.uniform(1e-6, 0.1),
            "noise_std": lambda spec: np.random.uniform(1e-6, 0.1),
            "l2coeff": lambda spec: np.random.uniform(1e-6, 0.1),
        },
    )
)

grid_search = (
    ml.Experiment()
    .components("optimization")
    .tune(
        stop={"accurary": 80},
        num_samples=100,
        config={"learning_rate": {"grid_search": [0.1, 0.05, 0.01]}},
    )
)



import os

from machinable.core.settings import get_settings


def test_settings(helpers):
    path = helpers.tmp_directory("settings")

    q = get_settings(file=os.path.join(path, "not-existing"))
    assert not q["cache"]["imports"]

    # empty file
    fp = os.path.join(path, "empty")
    open(fp, "w").close()
    q = get_settings(file=fp)
    assert not q["cache"]["imports"]



import tensorflow_datasets as tfds
from machinable import Mixin


class DataMixin(Mixin):
    def load_data(self):
        self.data = tfds.load(name=self.config.dataset.name).batch(
            self.config.dataset.batch_size
        )



import ast
import glob
import importlib
import inspect
import os
import re
import types


def get_file_list():
    path = os.path.dirname(__file__)
    return glob.glob(os.path.join(path, "_*.md"))


def read_file(path):
    with open(path, encoding="utf-8") as f:
        return f.read()


def parse_argumentblock(docstr, level="###"):
    regex = r"\n([^\W0-9]\w*):"

    def repl(m):
        variable = m.group(1)

        return "\n- **" + variable + "**:"

    docstr = re.sub(regex, repl, docstr)

    # highlight expr
    for expr in ("Integer", "String", "Boolean", "True", "False", "None"):
        docstr = docstr.replace(expr, "``" + expr + "``")

    docstr = level + " Arguments\n" + docstr

    return docstr


def parse_docstr(docstr, level="###", intent=4):
    if docstr is None:
        return ""

    # lower the headings
    docstr = docstr.replace((" " * 4) + "# ", "    " + level + " ")
    # remove intent
    docstr = docstr.replace("\n" + (" " * intent), "\n")
    # parse Arguments block
    regex = level + r" Arguments((?:(?:\n|\r\n?).+)+)"
    docstr = re.sub(regex, lambda m: parse_argumentblock(m.group(1), level), docstr)
    return docstr


def parse_method(obj, path, options, level="##"):
    method = getattr(obj, path)
    name = method.__name__.replace("_", "\_")
    signature = inspect.signature(method)
    md = level + " " + str(name) + "\n``" + path + str(signature) + "``\n\n"
    md += parse_docstr(method.__doc__, level=level + "#", intent=8)
    return md


def parse_property(obj, path, options, level="##"):
    p = getattr(obj, path)

    md = level + " " + path + ' <Badge text="property" vertical="middle" />\n\n'
    md += parse_docstr(p.__doc__, level=level + "#", intent=8)
    return md


def parse_class(obj, path, options):
    signature = inspect.signature(obj)
    level = options.get("level", "###")
    if options.get("header", True):
        md = "## " + obj.__name__ + "\n``" + path + str(signature) + "``\n\n"
    else:
        md = "``" + path + str(signature) + "``\n\n"
    md += parse_docstr(obj.__doc__, level=level + "#") + "\n"

    def filter_attributes(array):
        # filter hidden
        array = [m for m in array if not m.startswith("_")]
        # filter no-docs
        array = [m for m in array if getattr(obj, m).__doc__ is not None]

        if options.get("include", False):
            return [m for m in array if m in options["include"]]

        if options.get("exclude", False):
            return [m for m in array if m not in options["exclude"]]

        return array

    # find properties
    properties = [p for p in dir(obj) if isinstance(getattr(obj, p), property)]

    for p in sorted(filter_attributes(properties)):
        md += parse_property(obj, p, {}, level=level) + "\n"

    # find methods
    methods = [
        method_name
        for method_name in dir(obj)
        if callable(getattr(obj, method_name))
        and inspect.ismethod(getattr(obj, method_name))
        is False  # excludes class methods
    ]

    for method in sorted(filter_attributes(methods)):
        md += parse_method(obj, method, {}, level=level) + "\n"

    return md


def parse_module(obj, path, options):
    return parse_docstr(obj.__doc__)


def parse_function(obj, path, options, level="##"):
    signature = inspect.signature(obj)
    md = level + " " + obj.__name__ + "\n``" + path + str(signature) + "``\n\n"
    md += parse_docstr(obj.__doc__, level=level + "#")
    return md


def parse(filename):
    content = read_file(filename)

    regex = r"\{%\s+(.*)\s+%\}"

    def repl(m):
        target = m.group(1)

        try:
            target, options = target.split("::")
        except ValueError:
            options = "{}"

        options = ast.literal_eval(options)
        module, attribute = target.rsplit(".", 1)

        imp = importlib.import_module(module)
        obj = getattr(imp, attribute)
        if isinstance(obj, types.FunctionType):
            return parse_function(obj, target, options)
        if inspect.isclass(type(obj)):
            return parse_class(obj, target, options)
        elif inspect.ismodule(obj):
            return parse_module(obj, target, options)

        raise ValueError("Invalid object")

    md = re.sub(regex, repl, content)
    return md


def generate(filepath=None):
    if filepath is None:
        for fp in get_file_list():
            generate(fp)
        return

    directory, file = os.path.split(filepath)

    with open(
        os.path.join(
            os.path.dirname(os.path.dirname(directory)), "reference", file[1:]
        ),
        "w",
    ) as f:
        f.write(parse(filepath))


if __name__ == "__main__":
    generate()



import click

from ...project import Project
from ...project.manager import fetch_imports


@click.group(invoke_without_command=True)
@click.pass_context
def vendor(ctx):
    if ctx.invoked_subcommand is not None:
        return

    click.echo(ctx.get_help())


@vendor.command()
def fetch():
    project = Project()
    fetch_imports(project)
    click.echo("Dependencies have been fetched")



import asyncio

from .subscription_type import subscription


@subscription.source("observations")
async def observations_generator(obj, info):
    for i in range(5):
        await asyncio.sleep(1)
        yield str(i)


@subscription.field("observations")
def observations_resolver(log, info):
    return log



import copy
import random

from ..utils.dicts import update_dict
from ..utils.utils import generate_seed, generate_uid


def parse_experiment(specification, seed=None):
    seed_random_state = random.Random(seed)
    uid_random_state = random.Random(seed)

    # repeat behaviour
    repeats = []
    for level, repeat in enumerate(specification["repeats"]):
        # for each repeat that has been specified we proliferate
        k = repeat["arguments"]["k"]
        name = repeat["arguments"]["name"]
        # todo: support resuming mode that allows for nested cross validation etc
        #  mode = repeat['arguments']['mode']
        repeat_seed = generate_seed(seed_random_state)
        # collect repeat configuration and extend already existing inner repeats
        repeat_collection = []
        for r in range(k):
            suffix = "_" + str(level) if level > 0 else ""
            info = {
                name + "_SEED" + suffix: repeat_seed,
                name + "_TOTAL" + suffix: k,
                name + "_NUMBER" + suffix: r,
            }
            loop = [update_dict(inner, info, copy=True) for inner in repeats]
            if len(loop) == 0:
                loop = [info]
            repeat_collection.extend(loop)
        repeats = repeat_collection
    if not len(repeats):
        repeats.append({})

    for repeat in repeats:
        for component in specification["components"]:
            node_arguments = copy.deepcopy(component["arguments"])
            components = node_arguments.pop("components")
            if not isinstance(components, (list, tuple)):
                components = [components]
            resources = node_arguments.pop("resources", None)
            node = node_arguments.pop("node")

            node.flags["GLOBAL_SEED"] = seed
            node.flags["SEED"] = generate_seed(random_state=seed_random_state)
            node.flags["UID"] = generate_uid(random_state=uid_random_state)[0]
            node.flags.update(repeat)

            yield node, components, resources



import os

from fs import open_fs
from fs.base import FS
from fs.errors import CreateFailed as CreateFailedException
from orator import Model
from orator.orm import has_many


class Storage(Model):

    __guarded__ = ["id"]

    @has_many
    def tasks(self):
        from .task import Task

        return Task

    def filesystem(self):
        return open_fs(self.url)

    def available(self):
        try:
            return self.filesystem().exists(".")
        except CreateFailedException:
            return False

    @classmethod
    def get_url(cls, path):
        storage = os.path.expanduser(path)

        if "://" not in storage:
            storage = "osfs://" + storage

        return storage

    @classmethod
    def whereUrl(cls, storage):
        query = cls.query()
        return query.where("url", cls.get_url(storage))

    def index(self, filesystem=None):
        from .task import Task

        if filesystem is None:
            try:
                filesystem = self.filesystem()
            except CreateFailedException:
                filesystem = False

        # delete all stored experiment of this filesystem before re-indexing
        for task in Task.belongs_to_storage(self.id).get():
            # explicit such that delete event is triggered that cleans up the observations
            task.delete()

        # remove not available filesystem
        if not isinstance(filesystem, FS):
            self.delete()
            if isinstance(filesystem, CreateFailedException):
                raise filesystem
            else:
                raise CreateFailedException("Invalid filesystem: " + str(filesystem))

        for directory in filesystem.listdir(""):
            if len(directory) == 6:
                Task.createFromFile(filesystem, path=directory, storage_id=self.id)

            if directory == "_reruns":
                rerun = 1
                while filesystem.isdir("_reruns/" + str(rerun)):
                    # find subtasks
                    for rerun_task in filesystem.listdir("_reruns/" + str(rerun)):
                        if len(rerun_task) == 6:
                            Task.createFromFile(
                                filesystem,
                                path="_reruns/" + str(rerun) + "/" + rerun_task,
                                storage_id=self.id,
                            )
                    rerun += 1

        self.indexed = True
        self.save()
        return True


def deleting(storage):
    # clean up tasks
    from .task import Task

    Task.belongs_to_storage(storage.id).delete()


Storage.deleting(deleting)



import random
from typing import Optional

from baseconv import base62


def encode_experiment_id(seed, or_fail=True) -> Optional[str]:
    """Encodes a seed and returns the corresponding experiment ID

    # Arguments
    seed: int in the range 62^5 <= seed <= 62^6-1
    or_fail: If True, raises a Value error instead of returning None
    """
    try:
        if not isinstance(seed, int):
            raise ValueError()
        if 62 ** 5 <= seed <= 62 ** 6 - 1:
            return base62.encode(seed)
        raise ValueError()
    except (ValueError, TypeError):
        if or_fail:
            raise ValueError("Seed has to lie in range 62^5 <= seed <= 62^6-1")
        return None


def decode_experiment_id(experiment_id, or_fail=True) -> Optional[int]:
    """Decodes a experiment ID into the corresponding seed

    # Arguments
    experiment_id: The base62 experiment ID
    or_fail: If True, raises a Value error instead of returning None
    """
    try:
        if not isinstance(experiment_id, str):
            raise ValueError()
        value = int(base62.decode(experiment_id))
        if 62 ** 5 <= value <= 62 ** 6 - 1:
            return value
        raise ValueError()
    except (ValueError, TypeError):
        if or_fail:
            raise ValueError(f"'{experiment_id}' is not a valid experiment ID")
        return None


def generate_experiment_id(with_encoding=True, random_state=None):
    if random_state is None or isinstance(random_state, int):
        random_state = random.Random(random_state)

    # ~ 55x10^9 distinct experiment ids that if represented in base62 are len 6
    experiment_id = random_state.randint(62 ** 5, 62 ** 6 - 1)

    if with_encoding:
        return experiment_id, encode_experiment_id(experiment_id)

    return experiment_id



import tensorflow_datasets as tfds
from machinable import Component


class Classifier(Component):
    def load_data(self):
        self.data = tfds.load(name=self.config.dataset.name).batch(
            self.config.dataset.batch_size
        )

    def train(self):
        self.load_data()
        self.network.fit(self.data)



from typing import Mapping

from dotmap import DotMap


def get_or_fail(dict_likes, key, error="'{}' not found", options=None):
    if not isinstance(dict_likes, (list, tuple)):
        dict_likes = [dict_likes]

    for i, dict_like in enumerate(dict_likes):
        if key in dict_like:
            return dict_like[key]

    if isinstance(options, (list, tuple)):
        options = "\n Available: " + ", ".join(options)
    else:
        options = ""

    raise KeyError(error.format(key) + options)


def update_dict(d, update=None, copy=False):
    if d is None:
        d = {}
    if copy:
        d = d.copy()
    if not update or update is None:
        return d
    for k, v in update.items():
        if isinstance(v, Mapping):
            d[k] = update_dict(d.get(k, {}), v, copy=copy)
        else:
            d[k] = v
    return d


def read_path_dict(dict_like, path):
    """
    Resolves an non-recursive jsonpath like string over a given dictionary

    # Arguments
    dict_like: Mapping, Lookup dictionary
    path: String, lookup path, e.g. path.to[key].foo

    # Examples
    ```python
    read_path_dict({'foo': {'bar': 42}}, 'foo.bar')
    >>> 42
    ```

    # Raises
    KeyError
    """
    path = path.replace("][", "].[")
    segments = path.split(
        "."
    )  # todo: escaped_split(string=path, delimiter='.', escaper='\\')

    current = dict_like
    for segment in segments:

        if segment.find("[") >= 0:
            # parse array notation x[y] -> key: x, segment: y
            key = segment[segment.find("[") + 1 : -1]
            if key.isdigit():
                key = int(key)

            if not segment.startswith("["):
                current = current[segment[: segment.find("[")]]
            current = current[key]
        else:
            current = current[segment]

    return current


def serialize(obj):
    """JSON serializer for objects not serializable by default json code
    """
    if getattr(obj, "__dict__", False):
        return obj.__dict__

    return str(obj)


def dot_map(d=None):
    """Makes a dictionary into a dot accessible

    # Arguments
    d: dict-like

    Returns:
    ``dotmap.DotMap``
    """
    if d is None:
        d = {}
    return DotMap(d)



from machinable import Component


class LinearRegressionDummy(Component):
    def on_create(self):
        self.alpha = 0

    def train(self, x):
        self.alpha = 1 / x

    def on_destroy(self):
        print("Training finished")



from .core import Component, Mixin
from .engine import Engine
from .execution import Execution
from .experiment import Experiment
from .experiment import ExperimentComponent as C
from .project import Project
from .storage import Storage


def execute(experiment, storage=None, engine=None, project=None, seed=None):
    """Executes a machinable experiment

    Schedules the experiment for execution.

    # Arguments
    experiment: machinable.Experiment, specifies the execution experiment. For convenience, it can also be a string or tuple that
        defines the node argument, e.g. ``ml.execute('my_model')`` or ``ml.execute((('my_model', {'param': 1}),))``
    storage: String, URL of the write location for the results. If unspecified, the result write will be
        non-persistent. Remote locations like SSH or S3 are supported via
        [pyFilesystem URLs](https://pyfilesystem.readthedocs.io/en/latest/filesystems.html)
    engine: machinable.Engine|Dict|String|None, engine that handles execution,
        e.g. 'local' or 'ray' etc.
    project: Project|Dict|String|None, project used, defaults to current working directory
    seed: Integer|String|None, determines the global random seed. If None, a random seed will be generated.
        To re-use the same random seed of a previous execution, you can pass in its [experiment ID](.)

    # Example
    ```python
    import machinable as ml
    ml.execute(ml.Experiment().components('iris', 'random_forest'), 's3://bucket', seed=42)
    ```

    # Functional API
    You can use this method as a function decorator to implement a custom execution.
    The decorated function is invoked with the configuration as well as an
    store object, for example:

    ```python
    @ml.execute
    def custom_execute(component, components, store):
        store.log.info('Custom training with learning_rate=' + str(component.config.lr))

    custom_execute(experiment, storage, seed) # invokes the decorated function
    ```

    # Using the Execution

    This method forms a wrapper around machinable.Execution. You can instantiate
    machinable.Executon directly with the same argument to benefit from
    more fine grained execution APIs like asynchronous executon etc.

    # Returns

    The execution returns an machinable.Execution object that contains the
    result of the execution
    """
    if callable(experiment):
        # decorator use
        if None not in (storage, engine, project, seed):
            raise ValueError(
                "execute decorator takes no arguments; "
                "call the decorated function with arguments instead."
            )
        functional_component = experiment

        def wrapper(experiment, storage=None, engine=None, project=None, seed=None):
            project = Project.create(project)
            project.default_component = functional_component
            return (
                Execution(
                    experiment=experiment,
                    storage=storage,
                    engine=engine,
                    project=project,
                    seed=seed,
                )
                .summary()
                .submit()
            )

        return wrapper

    return (
        Execution(
            experiment=experiment,
            storage=storage,
            engine=engine,
            project=project,
            seed=seed,
        )
        .summary()
        .submit()
    )



import logging
import sys


class Log:
    """Logs to the console and the write

    ::: tip
    Becomes available as ``self.store.log`` and ``self.log``
    :::

    Exposes the standard interface of Python loggers, e.g. info, debug etc.

    # Arguments
    store: machinable.store.Observer, parent store instance
    config: dict, configuration options
    """

    def __init__(self, observer, config=None):
        self.store = observer
        self.config = config
        self.logger = self._get_logger()

    def _get_logger(self):
        logger = logging.getLogger(__name__)
        logger.handlers = []
        logger.propagate = False

        logger.setLevel(logging.INFO)

        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        ch.setFormatter(
            logging.Formatter(
                fmt=f"{self.store.directory()}; %(asctime)s; %(levelname)s: %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )
        logger.addHandler(ch)

        fileh = logging.StreamHandler(self.store.get_stream("log.txt", "a"))
        fileh.setFormatter(
            logging.Formatter(
                "%(asctime)s; %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
            )
        )

        logger.addHandler(fileh)

        return logger

    # forward function calls to logger

    def __getattr__(self, item):
        def forward(*args, **kwargs):
            method = getattr(self.logger, item)
            if hasattr(self.store, "events"):
                self.store.events.trigger("store.on_change", "log." + item)
            return method(*args, **kwargs)

        return forward



from machinable.project import Project
from machinable.project.manager import fetch_imports


def test_fetch_imports():
    project = Project("./test_project")
    assert fetch_imports(project) == ["fooba"]



import json

from ariadne import ScalarType

json_scalar = ScalarType("JSON")
datetime_scalar = ScalarType("DateTime")


@datetime_scalar.serializer
def serialize_datetime(value):
    # pendulum instance
    return value.isoformat()


@json_scalar.serializer
def serialize_json(value):
    if value is None:
        return None
    return json.dumps(value)


@json_scalar.value_parser
def parse_json(value):
    if value is None:
        return None
    return json.loads(value)


scalar_types = [datetime_scalar, json_scalar]



from machinable import Mixin


class TestMixin(Mixin):

    attribute = "works"

    def is_bound(self, param):
        return "bound_to_" + self.flags.BOUND + "_" + str(param)

    def this_reference(self, param):
        return self.__mixin__.is_bound("and_referenced_" + str(param))

    def this_attribute(self):
        return self.__mixin__.attribute

    def this_static(self, param):
        return self.__mixin__.static_method(param)

    @staticmethod
    def static_method(foo):
        return foo

    @property
    def key_propery(self):
        return 1



import copy
import importlib
import inspect
import mimetypes
import os
import pickle
import sys

import gitignore_parser
from fs.zipfs import WriteZipFS

from machinable.utils.utils import is_valid_variable_name

from ..config.loader import from_callable as load_config_from_callable
from ..config.loader import from_file as load_config_file
from ..config.loader import from_string as load_config_from_string
from ..config.parser import parse_module_list
from ..core import Component as BaseComponent
from ..core import Mixin as BaseMixin
from ..core.component import FunctionalComponent
from ..core.settings import get_settings
from ..utils.dicts import update_dict
from ..utils.formatting import exception_to_str, msg
from ..utils.traits import Jsonable
from ..utils.vcs import get_commit
from .manager import fetch_imports
from .registration import Registration


class Project(Jsonable):
    def __init__(self, options=None, parent=None):
        if callable(options):
            if not inspect.isclass(options):
                options = FunctionalComponent(options)
            options = {"default_component": options}
        if isinstance(options, str):
            options = {"directory": options}

        self.options = update_dict(
            {
                "directory": os.getcwd(),
                "config_file": "machinable.yaml",
                "config": None,
                "registration_module": "_machinable",
                "vendor_caching": get_settings()["cache"].get("imports", False),
                "default_component": None,
            },
            options,
        )
        self.parent = parent

        self.config = None
        self.parsed_config = None
        self._registration = None

        if (
            os.path.exists(self.directory_path)
            and self.is_root()
            and self.directory_path not in sys.path
        ):
            sys.path.insert(0, self.directory_path)

    def __repr__(self):
        return f"Project({self.directory_path})"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def create(cls, args=None):
        """Creates a project instance"""
        if isinstance(args, cls):
            return args

        if args is None:
            return cls()

        if callable(args):
            return cls({"default_component": args})

        if isinstance(args, str):
            return cls({"directory": args})

        if isinstance(args, tuple):
            return cls(*args)

        raise ValueError(f"Invalid project: {args}")

    @property
    def config_filepath(self):
        return os.path.join(self.options["directory"], self.options["config_file"])

    @property
    def directory_path(self):
        return os.path.abspath(self.options["directory"])

    @property
    def directory_prefix(self):
        if self.options["directory"] == ".":
            return ""

        return self.options["directory"]

    @property
    def vendor_directory(self):
        return os.path.join(self.options["directory"], "vendor")

    @property
    def vendor_path(self):
        return os.path.abspath(self.vendor_directory)

    @property
    def import_prefix(self):
        # find relative address to target import class
        prefix = os.path.relpath(self.directory_path, os.getcwd())

        if prefix == ".":
            return None

        # todo: support for relative directory that contain ../

        return prefix.replace("/", ".")

    @property
    def default_component(self):
        return self.options["default_component"]

    @default_component.setter
    def default_component(self, value):
        if callable(value) and not inspect.isclass(value):
            value = FunctionalComponent(value)

        self.options["default_component"] = value

    def serialize(self):
        return copy.deepcopy(self.options)

    @classmethod
    def unserialize(cls, serialized):
        return cls(serialized)

    def is_root(self):
        return self.parent is None

    def get_root(self):
        if self.is_root():
            return self

        p = self.parent
        while not p.is_root():
            p = p.parent

        return p

    def has_config_file(self):
        return os.path.isfile(
            os.path.join(self.directory_path, self.options["config_file"])
        )

    def has_registration(self):
        return os.path.isfile(
            os.path.join(
                self.directory_path, self.options["registration_module"] + ".py"
            )
        ) or os.path.isfile(
            os.path.join(
                self.directory_path, self.options["registration_module"], "__init__.py"
            )
        )

    def exists(self):
        return os.path.isfile(self.directory_path)

    @property
    def registration(self):
        if self._registration is None:
            if not self.has_registration():
                self._registration = Registration()
            else:
                try:
                    registration_module = importlib.import_module(
                        self.options["registration_module"]
                    )
                except ImportError as ex:
                    msg(
                        f"Could not import project registration. {ex}\n{exception_to_str(ex)}",
                        level="error",
                        color="fail",
                    )

                registration_class = getattr(registration_module, "Project", False)
                if not registration_class:
                    registration_class = Registration

                self._registration = registration_class()

        return self._registration

    def get_config(self, cached="auto"):
        if cached == "auto":
            # todo: reliably compute hash including vendors to determine changes
            cashed = False

        if cached is not False and self.config is not None:
            return self.config

        # load config
        self.parsed_config = None
        if self.options["config"] is not None:
            # direct specification
            if isinstance(self.options["config"], str):
                self.config = load_config_from_string(self.options["config"])
            else:
                self.config = self.options["config"]
            if not isinstance(self.config, dict):
                raise ValueError(
                    f"Invalid project configuration: '{self.options['options']}'"
                )
        # elif isinstance(self.default_component, FunctionalComponent):
        #     config = load_config_from_callable(self.default_component.function)
        #     if config is None:
        #         config = load_config_file(self.config_filepath, default={})
        #     self.config = config
        elif self.has_config_file():
            self.config = load_config_file(self.config_filepath, default={})
        else:
            self.config = {}

        # validate and set defaults
        self.config.setdefault("_evaluate", True)

        self.config.setdefault("+", [])
        assert isinstance(
            self.config["+"], (tuple, list)
        ), "Invalid import definition: " + str(self.config["+"])

        self.config.setdefault("mixins", [])

        return self.config

    def backup_source_code(self, filepath="code.zip", opener=None) -> bool:
        """Writes all files in project (excluding those in .gitignore) to zip file

        # Arguments
        filepath: String, target file
        opener: Optional file opener object. Defaults to built-in `open`

        # Returns
        True on completion
        """
        if opener is None:
            opener = open

        # Get stream in the PyFS to create zip file with
        zip_stream = opener(filepath, "wb")

        # Get object to test whether files are in .gitignore
        try:
            gitignore = gitignore_parser.parse_gitignore(
                os.path.join(self.directory_prefix, ".gitignore")
            )
        except FileNotFoundError:

            def gitignore(*args, **kwargs):
                return False

        with WriteZipFS(file=zip_stream) as zip_fs:
            for folder, subfolders, filenames in os.walk(self.directory_path):

                # Remove subfolders that are listed in .gitignore, so they won't be explored later
                subfolders[:] = [
                    sub
                    for sub in subfolders
                    if sub != ".git" and not gitignore(os.path.join(folder, sub))
                ]

                for file_name in filenames:
                    # Find absolute path of file for .gitignore checking
                    abspath_file = os.path.join(folder, file_name)

                    # Ignore binary file types
                    file_type = mimetypes.guess_type(abspath_file)[0]
                    if (
                        file_type is None or not file_type.startswith("text/")
                    ) and not file_name.endswith(".yaml"):
                        continue

                    if gitignore(abspath_file):
                        # Skip file if it is listed in .gitignore
                        continue

                    # Get relative paths (which will be replicated in zip)
                    relpath_folder = os.path.relpath(folder, self.directory_path)
                    relpath_file = os.path.join(relpath_folder, file_name)

                    # Make directories in zip if necessary
                    if not zip_fs.exists(relpath_folder):
                        # Need to use makedirs (not makedir) in case file in nested folders
                        zip_fs.makedirs(relpath_folder)

                    # Read file contents
                    file_content = open(
                        os.path.join(self.directory_prefix, relpath_file), "r"
                    ).read()

                    # Add file to zip
                    zip_fs.writetext(relpath_file, file_content)

            # Add zip to PyFS
            zip_fs.write_zip()

        return True

    def get_code_version(self):
        return {
            "project": get_commit(self.config_filepath),
            "vendor": [
                get_commit(
                    os.path.join(
                        self.directory_path, "vendor", vendor, "machinable.yaml"
                    )
                )
                for vendor in [next(iter(d)) for d in self.get_config()["+"]]
            ],
        }

    def reload_imports(self):
        vendor_caching = self.options["vendor_caching"]
        self.options["vendor_caching"] = False
        self.parse_imports()
        self.options["vendor_caching"] = vendor_caching

    def parse_imports(self, cached=None):
        if cached is None:
            cached = self.get_root().options["vendor_caching"]

        config = self.get_config()["+"]

        if len(config) == 0:
            return {"components": None, "mixins": None}

        import_names = fetch_imports(self, config)

        imports = {"components": {}, "mixins": {}}

        for import_name in import_names:
            # use pickled import cache if existing
            import_cache = os.path.join(
                self.directory_path, "vendor", ".cache", import_name + ".p"
            )
            if cached and os.path.isfile(import_cache):
                with open(import_cache, "rb") as f:
                    import_config = pickle.load(f)
            else:
                # use project abstraction to resolve imports
                import_project = Project(
                    os.path.join(self.options["directory"], "vendor", import_name),
                    parent=self,
                )

                if not import_project.has_config_file():
                    continue

                import_config = import_project.parse_config()

                # write pickle-cache
                os.makedirs(os.path.dirname(import_cache), exist_ok=True)
                with open(import_cache, "wb") as f:
                    pickle.dump(import_config, f)

            # merge into global imports collection
            for section in ("components", "mixins"):
                for k, v in import_config[section].items():
                    imports[section][import_name + "." + k] = v

        return imports

    def parse_config(self, reference=True, cached="auto"):
        if cached == "auto":
            # todo: detect from hash
            cached = False

        if not cached and self.parsed_config is not None:
            return self.parsed_config

        config = self.get_config(cached)

        imports = self.parse_imports()

        if reference:
            # pass in the full config as $ root reference
            reference = config

        collections = [k for k in config.keys() if k.startswith("components")]
        components = {}
        top_level_alias = {}
        for collection in collections:
            if config[collection] is None:
                continue

            scope = collection.replace("components:", "")
            if scope.find("=") != -1:
                scope, alias = scope.split("=")
                if alias in top_level_alias:
                    raise ValueError(
                        f"Alias '{alias}' of 'components:{collection}' is ambiguous"
                    )
            else:
                alias = scope
                if alias in top_level_alias:
                    raise ValueError(f"'components:{collection}' already exists")

            if not is_valid_variable_name(alias):
                raise ValueError(
                    f"Alias '{alias}' of 'components:{collection}' is not a valid Python variable name"
                )

            top_level_alias[alias] = collection

            parse_module_list(
                config[collection],
                scope=scope if scope != "components" else None,
                baseclass=BaseComponent,
                modules=components,
                imports=imports["components"],
                reference=reference,
                import_prefix=self.import_prefix,
                auto_alias=scope if scope != "components" else None,
            )

        mixins = parse_module_list(
            config.get("mixins"),
            scope=None,
            baseclass=BaseMixin,
            imports=[imports["mixins"], imports["components"]],
            reference=reference,
            import_prefix=self.import_prefix,
            auto_alias="mixins",
        )

        self.parsed_config = {
            "imports": imports,
            "mixins": mixins,
            "components": components,
            # globals
            "_evaluate": config["_evaluate"],
        }

        return self.parsed_config



from machinable.config.loader import from_callable, from_file


def test_config_loader_from_file():
    data = from_file("./test_project/machinable.yaml")

    # outsourcing $/
    assert data["outsource"]["hello"] == "success"

    # correct scientific notation parsing
    assert data["scientific"] == 5e-6


def test_config_loader_from_callable():
    def test_callable():
        """Normal doc string No machinable config here"""
        pass

    def test_machinable_callable():
        """Normal doc string
        # machinable.yaml
        a: 1
        nested:
          level: 2
        """
        pass

    assert from_callable(test_callable) is None

    config = from_callable(test_machinable_callable)
    assert config["a"] == 1
    assert config["nested"]["level"] == 2



from orator import Schema


def run(db):
    schema = Schema(db)

    schema.drop_if_exists("storages")
    with schema.create("storages") as table:
        table.increments("id")
        table.string("url").nullable()
        table.boolean("indexed").default(False)
        table.timestamps()

    schema.drop_if_exists("tasks")
    with schema.create("tasks") as table:
        table.increments("id")
        table.string("name").nullable()
        table.string("task_id").nullable()
        table.string("execution_id").nullable()
        table.integer("seed").nullable()
        table.boolean("tune").nullable()
        table.integer("execution_cardinality").nullable()
        table.integer("rerun").nullable()
        table.boolean("code_backup").nullable()
        table.text("code_version").nullable()
        table.string("path").nullable()
        table.integer("storage_id").nullable()
        table.integer("observations_count").nullable()
        table.timestamp("started").nullable()
        table.timestamp("heartbeat").nullable()
        table.timestamp("finished").nullable()
        table.timestamps()

    schema.drop_if_exists("observations")
    with schema.create("observations") as table:
        table.increments("id")
        table.string("path").nullable()
        table.string("node").nullable()
        table.integer("execution_index").nullable()
        table.string("components").nullable()
        table.timestamp("started").nullable()
        table.timestamp("heartbeat").nullable()
        table.timestamp("finished").nullable()
        table.integer("task_id").nullable()
        table.timestamps()

    return True



# This file contains modified 3rd party source code from
# https://github.com/sdispater/backpack/blob/master/tests/collections/test_collection.py
# The copyright and license agreement can be found in the ThirdPartyNotices.txt file at the root of this repository.


from unittest import TestCase

from machinable.v1.observations import collect
from machinable.v1.observations.collections import Collection


def test_collect():
    assert isinstance(collect([1, 2]), Collection)


class CollectionTestCase(TestCase):
    def test_first_returns_first_item_in_collection(self):
        c = Collection(["foo", "bar"])

        self.assertEqual("foo", c.first())

    def test_last_returns_last_item_in_collection(self):
        c = Collection(["foo", "bar"])

        self.assertEqual("bar", c.last())

    def test_pop_removes_and_returns_last_item_or_specified_index(self):
        c = Collection(["foo", "bar"])

        self.assertEqual("bar", c.pop())
        self.assertEqual("foo", c.last())

        c = Collection(["foo", "bar"])

        self.assertEqual("foo", c.pop(0))
        self.assertEqual("bar", c.first())

    def test_empty_collection_is_empty(self):
        c = Collection()
        c2 = Collection([])

        self.assertTrue(c.empty())
        self.assertTrue(c2.empty())

    def test_collection_is_constructed(self):
        c = Collection("foo")
        self.assertEqual(["foo"], c.all())

        c = Collection(2)
        self.assertEqual([2], c.all())

        c = Collection(False)
        self.assertEqual([False], c.all())

        c = Collection(None)
        self.assertEqual([], c.all())

        c = Collection()
        self.assertEqual([], c.all())

    def test_offset_access(self):
        c = Collection(["foo", "bar"])
        self.assertEqual("bar", c[1])

        c[1] = "baz"
        self.assertEqual("baz", c[1])

        del c[0]
        self.assertEqual("baz", c[0])

    def test_forget(self):
        c = Collection(["foo", "bar", "boom"])
        c.forget(0)
        self.assertEqual("bar", c[0])
        c.forget(0, 1)
        self.assertTrue(c.empty())

    def test_get_avg_items_from_collection(self):
        c = Collection([{"foo": 10}, {"foo": 20}])
        self.assertEqual(15, c.avg("foo"))

        c = Collection([1, 2, 3, 4, 5])
        self.assertEqual(3, c.avg())

        c = Collection()
        self.assertIsNone(c.avg())

    def test_collapse(self):
        obj1 = object()
        obj2 = object()

        c = Collection([[obj1], [obj2]])
        self.assertEqual([obj1, obj2], c.collapse().all())

    def test_collapse_with_nested_collection(self):
        c = Collection([Collection([1, 2, 3]), Collection([4, 5, 6])])
        self.assertEqual([1, 2, 3, 4, 5, 6], c.collapse().all())

    def test_contains(self):
        c = Collection([1, 3, 5])

        self.assertTrue(c.contains(1))
        self.assertFalse(c.contains(2))
        self.assertTrue(c.contains(lambda x: x < 5))
        self.assertFalse(c.contains(lambda x: x > 5))
        self.assertIn(3, c)

        c = Collection([{"v": 1}, {"v": 3}, {"v": 5}])
        self.assertTrue(c.contains("v", 1))
        self.assertFalse(c.contains("v", 2))

        obj1 = type("lamdbaobject", (object,), {})()
        obj1.v = 1
        obj2 = type("lamdbaobject", (object,), {})()
        obj2.v = 3
        obj3 = type("lamdbaobject", (object,), {})()
        obj3.v = 5
        c = Collection([{"v": 1}, {"v": 3}, {"v": 5}])
        self.assertTrue(c.contains("v", 1))
        self.assertFalse(c.contains("v", 2))

    def test_countable(self):
        c = Collection(["foo", "bar"])
        self.assertEqual(2, c.count())
        self.assertEqual(2, len(c))

    def test_diff(self):
        c = Collection(["foo", "bar"])
        self.assertEqual(["foo"], c.diff(Collection(["bar", "baz"])).all())

    def test_each(self):
        original = ["foo", "bar", "baz"]
        c = Collection(original)

        result = []
        c.each(lambda x: result.append(x))
        self.assertEqual(result, original)
        self.assertEqual(original, c.all())

    def test_every(self):
        c = Collection([1, 2, 3, 4, 5, 6])
        self.assertEqual([1, 3, 5], c.every(2).all())
        self.assertEqual([2, 4, 6], c.every(2, 1).all())

    def test_filter(self):
        c = Collection([{"id": 1, "name": "hello"}, {"id": 2, "name": "world"}])
        self.assertEqual(
            [{"id": 2, "name": "world"}], c.filter(lambda item: item["id"] == 2).all()
        )

        c = Collection(["", "hello", "", "world"])
        self.assertEqual(["hello", "world"], c.filter().all())

    def test_where(self):
        c = Collection([{"v": 1}, {"v": 3}, {"v": 2}, {"v": 3}, {"v": 4}])
        self.assertEqual([{"v": 3}, {"v": 3}], c.where("v", 3).all())

    def test_implode(self):
        obj1 = type("lamdbaobject", (object,), {})()
        obj1.name = "john"
        obj1.email = "foo"
        c = Collection(
            [{"name": "john", "email": "foo"}, {"name": "jane", "email": "bar"}]
        )
        self.assertEqual("foobar", c.implode("email"))
        self.assertEqual("foo,bar", c.implode("email", ","))

        c = Collection(["foo", "bar"])
        self.assertEqual("foobar", c.implode(""))
        self.assertEqual("foo,bar", c.implode(","))

    def test_map(self):
        c = Collection([1, 2, 3, 4, 5])
        self.assertEqual([3, 4, 5, 6, 7], c.map(lambda x: x + 2).all())

    def test_merge(self):
        c = Collection([1, 2, 3])
        c.merge([4, 5, 6])
        self.assertEqual([1, 2, 3, 4, 5, 6], c.all())

        c = Collection(Collection([1, 2, 3]))
        c.merge([4, 5, 6])
        self.assertEqual([1, 2, 3, 4, 5, 6], c.all())

    def test_prepend(self):
        c = Collection([4, 5, 6])
        c.prepend(3)
        self.assertEqual([3, 4, 5, 6], c.all())

    def test_append(self):
        c = Collection([3, 4, 5])
        c.append(6)
        self.assertEqual([3, 4, 5, 6], c.all())

    def test_pull(self):
        c = Collection([1, 2, 3, 4])
        c.pull(2)
        self.assertEqual([1, 2, 4], c.all())

    def test_put(self):
        c = Collection([1, 2, 4])
        c.put(2, 3)
        self.assertEqual([1, 2, 3], c.all())

    def test_reject(self):
        c = Collection([1, 2, 3, 4, 5, 6])
        self.assertEqual([1, 2, 3], c.reject(lambda x: x > 3).all())

    def test_reverse(self):
        c = Collection([1, 2, 3, 4])
        self.assertEqual([4, 3, 2, 1], c.reverse().all())

    def test_sort(self):
        c = Collection([5, 3, 1, 2, 4])

        sorted = c.sort(lambda x: x)
        self.assertEqual([1, 2, 3, 4, 5], sorted.all())

    def test_take(self):
        c = Collection([1, 2, 3, 4, 5, 6])
        self.assertEqual([1, 2, 3], c.take(3).all())
        self.assertEqual([4, 5, 6], c.take(-3).all())

    def test_transform(self):
        c = Collection([1, 2, 3, 4])
        c.transform(lambda x: x + 2)
        self.assertEqual([3, 4, 5, 6], c.all())

    def test_zip(self):
        c = Collection([1, 2, 3])
        self.assertEqual([(1, 4), (2, 5), (3, 6)], c.zip([4, 5, 6]).all())

    def test_only(self):
        c = Collection([1, 2, 3, 4, 5])
        self.assertEqual([2, 4], c.only(1, 3).all())

    def test_without(self):
        c = Collection([1, 2, 3, 4, 5])
        self.assertEqual([1, 3, 5], c.without(1, 3).all())
        self.assertEqual([1, 2, 3, 4, 5], c.all())

    def test_flatten(self):
        c = Collection({"foo": [5, 6], "bar": 7, "baz": {"boom": [1, 2, 3, 4]}})

        self.assertEqual([1, 2, 3, 4, 5, 6, 7], c.flatten().sort().all())

        c = Collection([1, [2, 3], 4])
        self.assertEqual([1, 2, 3, 4], c.flatten().all())



# This file contains modified 3rd party source code from
# https://github.com/drgrib/ConfigMap/blob/master/ConfigMap/test.py
# The copyright and license agreement can be found in the ThirdPartyNotices.txt file at the root of this repository.

import unittest

from machinable.config.mapping import ConfigMap


def test_evaluate():
    m = ConfigMap(
        {
            "test": 1,
            "_hidden": "invisible",
            "_mixin_": True,
            "~version": {"should be": "hidden"},
        }
    ).toDict(with_hidden=False)

    assert "_hidden" not in m
    assert m["_mixin_"]

    print(m)


def test_versions():
    m = ConfigMap(
        {
            "test": 1,
            "~a": {"a": "1"},
            "~b": {"b": "1", "~d": {"d": 1}},
            "~c": {"c": "1"},
        }
    )

    v = m.get_versioning()
    assert len(v) == 3
    assert "a" in v
    assert m["~b"].get_versioning()[0] == "d"


class TestReadme(unittest.TestCase):
    def test_basic_use(self):
        m = ConfigMap()
        self.assertIsInstance(m, ConfigMap)
        m.name = "Joe"
        self.assertEqual(m.name, "Joe")
        self.assertEqual("Hello " + m.name, "Hello Joe")
        self.assertIsInstance(m, dict)
        self.assertTrue(issubclass(m.__class__, dict))
        self.assertEqual(m["name"], "Joe")
        m.name += " Smith"
        m["name"] += " Jr"
        self.assertEqual(m.name, "Joe Smith Jr")

    def test_automatic_hierarchy(self):
        m = ConfigMap()
        m.people.steve.age = 31
        self.assertEqual(m.people.steve.age, 31)

    def test_key_init(self):
        m = ConfigMap(a=1, b=2)
        self.assertEqual(m.a, 1)
        self.assertEqual(m.b, 2)

    def test_dict_conversion(self):
        d = {"a": 1, "b": 2, "c": {"d": 3, "e": 4}}
        m = ConfigMap(d)
        self.assertEqual(m.a, 1)
        self.assertEqual(m.b, 2)
        d2 = m.toDict()
        self.assertIsInstance(d2, dict)
        self.assertNotIsInstance(d2, ConfigMap)
        self.assertEqual(len(d2), 3)
        self.assertEqual(d2["a"], 1)
        self.assertEqual(d2["b"], 2)
        self.assertNotIsInstance(d2["c"], ConfigMap)
        self.assertEqual(len(d2["c"]), 2)
        self.assertEqual(d2["c"]["d"], 3)
        self.assertEqual(d2["c"]["e"], 4)

    def test_ordered_iteration(self):
        m = ConfigMap()
        m.people.john.age = 32
        m.people.john.job = "programmer"
        m.people.mary.age = 24
        m.people.mary.job = "designer"
        m.people.dave.age = 55
        m.people.dave.job = "manager"
        expected = [
            ("john", 32, "programmer"),
            ("mary", 24, "designer"),
            ("dave", 55, "manager"),
        ]
        for i, (k, v) in enumerate(m.people.items()):
            self.assertEqual(expected[i][0], k)
            self.assertEqual(expected[i][1], v.age)
            self.assertEqual(expected[i][2], v.job)


class TestBasic(unittest.TestCase):
    def setUp(self):
        self.d = {"a": 1, "b": 2, "subD": {"c": 3, "d": 4}}

    def test_dict_init(self):
        m = ConfigMap(self.d)
        self.assertIsInstance(m, ConfigMap)
        self.assertEqual(m.a, 1)
        self.assertEqual(m.b, 2)
        self.assertIsInstance(m.subD, ConfigMap)
        self.assertEqual(m.subD.c, 3)
        self.assertEqual(m.subD.d, 4)

    def test_copy(self):
        m = ConfigMap(self.d)
        dm_copy = m.copy()
        self.assertIsInstance(dm_copy, ConfigMap)
        self.assertEqual(dm_copy.a, 1)
        self.assertEqual(dm_copy.b, 2)
        self.assertIsInstance(dm_copy.subD, ConfigMap)
        self.assertEqual(dm_copy.subD.c, 3)
        self.assertEqual(dm_copy.subD.d, 4)

    def test_fromkeys(self):
        m = ConfigMap.fromkeys([1, 2, 3], "a")
        self.assertEqual(len(m), 3)
        self.assertEqual(m[1], "a")
        self.assertEqual(m[2], "a")
        self.assertEqual(m[3], "a")

    def test_dict_functionality(self):
        m = ConfigMap(self.d)
        self.assertEqual(m.get("a"), 1)
        self.assertEqual(m.get("f", 33), 33)
        self.assertIsNone(m.get("f"))
        m.update([("rat", 5), ("bum", 4)], dog=7, cat=9)
        self.assertEqual(m.rat, 5)
        self.assertEqual(m.bum, 4)
        self.assertEqual(m.dog, 7)
        self.assertEqual(m.cat, 9)
        m.update({"lol": 1, "ba": 2})
        self.assertEqual(m.lol, 1)
        self.assertEqual(m.ba, 2)
        self.assertTrue("a" in m)
        self.assertFalse("c" in m)
        self.assertTrue("c" in m.subD)
        self.assertTrue(len(m.subD), 2)
        del m.subD.c
        self.assertFalse("c" in m.subD)
        self.assertTrue(len(m.subD), 1)

    def test_list_comprehension(self):
        parentDict = {
            "name": "Father1",
            "offspring": [{"name": "Child1"}, {"name": "Child2"}, {"name": "Child3"},],
        }
        parent = ConfigMap(parentDict)
        ordered_names = ["Child1", "Child2", "Child3"]
        comp = [x.name for x in parent.offspring]
        self.assertEqual(ordered_names, comp)


class TestPickle(unittest.TestCase):
    def setUp(self):
        self.d = {"a": 1, "b": 2, "subD": {"c": 3, "d": 4}}

    def test(self):
        import pickle

        pm = ConfigMap(self.d)
        s = pickle.dumps(pm)
        m = pickle.loads(s)
        self.assertIsInstance(m, ConfigMap)
        self.assertEqual(m.a, 1)
        self.assertEqual(m.b, 2)
        self.assertIsInstance(m.subD, ConfigMap)
        self.assertEqual(m.subD.c, 3)
        self.assertEqual(m.subD.d, 4)


class TestEmpty(unittest.TestCase):
    def test(self):
        m = ConfigMap()
        self.assertTrue(m.empty())
        m.a = 1
        self.assertFalse(m.empty())
        self.assertTrue(m.b.empty())
        self.assertIsInstance(m.b, ConfigMap)


class TestDynamic(unittest.TestCase):
    def test(self):
        m = ConfigMap()
        m.still.works
        m.sub.still.works
        nonDynamic = ConfigMap(_dynamic=False)

        def assignNonDynamic():
            nonDynamic.no

        self.assertRaises(KeyError, assignNonDynamic)

        nonDynamicWithInit = ConfigMap(m, _dynamic=False)
        nonDynamicWithInit.still.works
        nonDynamicWithInit.sub.still.works

        def assignNonDynamicWithInit():
            nonDynamicWithInit.no.creation

        self.assertRaises(KeyError, assignNonDynamicWithInit)


class TestRecursive(unittest.TestCase):
    def test(self):
        m = ConfigMap()
        m.a = 5
        m_id = id(m)
        m.recursive = m
        self.assertEqual(id(m.recursive.recursive.recursive), m_id)
        outStr = str(m)
        self.assertIn("""a=5""", outStr)
        self.assertIn("""recursive=ConfigMap(...)""", outStr)
        d = m.toDict()
        d_id = id(d)
        d["a"] = 5
        d["recursive"] = d
        d["recursive"]["recursive"]["recursive"]
        self.assertEqual(id(d["recursive"]["recursive"]["recursive"]), d_id)
        outStr = str(d)
        self.assertIn("""'a': 5""", outStr)
        self.assertIn("""recursive': {...}""", outStr)
        m2 = ConfigMap(d)
        m2_id = id(m2)
        self.assertEqual(id(m2.recursive.recursive.recursive), m2_id)
        outStr2 = str(m2)
        self.assertIn("""a=5""", outStr2)
        self.assertIn("""recursive=ConfigMap(...)""", outStr2)


class Testkwarg(unittest.TestCase):
    def test(self):
        a = {"1": "a", "2": "b"}
        b = ConfigMap(a, _dynamic=False)

        def capture(**kwargs):
            return kwargs

        self.assertEqual(a, capture(**b.toDict()))


class TestDeepCopy(unittest.TestCase):
    def test(self):
        import copy

        original = ConfigMap()
        original.a = 1
        original.b = 3
        shallowCopy = original
        deepCopy = copy.deepcopy(original)
        self.assertEqual(original, shallowCopy)
        self.assertEqual(id(original), id(shallowCopy))
        self.assertEqual(original, deepCopy)
        self.assertNotEqual(id(original), id(deepCopy))
        original.a = 2
        self.assertEqual(original, shallowCopy)
        self.assertNotEqual(original, deepCopy)

    def test_order_preserved(self):
        import copy

        original = ConfigMap()
        original.a = 1
        original.b = 2
        original.c = 3
        deepCopy = copy.deepcopy(original)
        orderedPairs = []
        for k, v in original.iteritems():
            orderedPairs.append((k, v))
        for i, (k, v) in enumerate(deepCopy.iteritems()):
            self.assertEqual(k, orderedPairs[i][0])
            self.assertEqual(v, orderedPairs[i][1])


class TestConfigMapTupleToDict(unittest.TestCase):
    def test(self):
        m = ConfigMap({"a": 1, "b": (11, 22, ConfigMap({"c": 3}))})
        d = m.toDict()
        self.assertEqual(d, {"a": 1, "b": (11, 22, {"c": 3})})


class TestOrderedDictInit(unittest.TestCase):
    def test(self):
        from collections import OrderedDict

        o = OrderedDict([("a", 1), ("b", 2), ("c", [OrderedDict([("d", 3)])])])
        m = ConfigMap(o)
        self.assertIsInstance(m, ConfigMap)
        self.assertIsInstance(m.c[0], ConfigMap)


class TestEmptyAdd(unittest.TestCase):
    def test_base(self):
        m = ConfigMap()
        for i in range(7):
            m.counter += 1
        self.assertNotIsInstance(m.counter, ConfigMap)
        self.assertIsInstance(m.counter, int)
        self.assertEqual(m.counter, 7)

    def test_various(self):
        m = ConfigMap()
        m.a.label = "test"
        m.a.counter += 2
        self.assertIsInstance(m.a, ConfigMap)
        self.assertEqual(m.a.label, "test")
        self.assertNotIsInstance(m.a.counter, ConfigMap)
        self.assertIsInstance(m.a.counter, int)
        self.assertEqual(m.a.counter, 2)
        m.a.counter += 1
        self.assertEqual(m.a.counter, 3)

    def test_proposal(self):
        my_counters = ConfigMap()
        pages = [
            "once upon a time",
            "there was like this super awesome prince",
            "and there was this super rad princess",
            "and they had a mutually respectful, egalitarian relationship",
            "the end",
        ]
        for stuff in pages:
            my_counters.page += 1
        self.assertIsInstance(my_counters, ConfigMap)
        self.assertNotIsInstance(my_counters.page, ConfigMap)
        self.assertIsInstance(my_counters.page, int)
        self.assertEqual(my_counters.page, 5)

    def test_string_addition(self):
        m = ConfigMap()
        m.quote += "lions"
        m.quote += " and tigers"
        m.quote += " and bears"
        m.quote += ", oh my"
        self.assertEqual(m.quote, "lions and tigers and bears, oh my")

    def test_strange_addition(self):
        m = ConfigMap()
        m += "I'm a string now"
        self.assertIsInstance(m, str)
        self.assertNotIsInstance(m, ConfigMap)
        self.assertEqual(m, "I'm a string now")
        m2 = ConfigMap() + "I'll replace that ConfigMap"
        self.assertEqual(m2, "I'll replace that ConfigMap")

    def test_protected_hierarchy(self):
        m = ConfigMap()
        m.protected_parent.key = "value"

        def protectedFromAddition():
            m.protected_parent += 1

        self.assertRaises(TypeError, protectedFromAddition)

    def test_type_error_raised(self):
        m = ConfigMap()

        def badAddition():
            m.a += 1
            m.a += " and tigers"

        self.assertRaises(TypeError, badAddition)


# Test classes for SubclassTestCase below

# class that overrides __getitem__
class MyConfigMap(ConfigMap):
    def __getitem__(self, k):
        return super(MyConfigMap, self).__getitem__(k)


# subclass with existing property
class PropertyConfigMap(MyConfigMap):
    def __init__(self, *args, **kwargs):
        super(MyConfigMap, self).__init__(*args, **kwargs)
        self._myprop = None

    @property
    def my_prop(self):
        if not self._myprop:
            self._myprop = PropertyConfigMap({"nested_prop": 123})
        return self._myprop


class SubclassTestCase(unittest.TestCase):
    def test_nested_subclass(self):
        my = MyConfigMap()
        my.x.y.z = 123
        self.assertEqual(my.x.y.z, 123)
        self.assertIsInstance(my.x, MyConfigMap)
        self.assertIsInstance(my.x.y, MyConfigMap)

    def test_subclass_with_property(self):
        p = PropertyConfigMap()
        self.assertIsInstance(p.my_prop, PropertyConfigMap)
        self.assertEqual(p.my_prop.nested_prop, 123)
        p.my_prop.second.third = 456
        self.assertIsInstance(p.my_prop.second, PropertyConfigMap)
        self.assertEqual(p.my_prop.second.third, 456)



import ast
import importlib
import inspect
import json
import os
import re
import shutil
from collections import namedtuple

from . import standalone


def copy_and_overwrite(from_path, to_path):
    if os.path.exists(to_path):
        shutil.rmtree(to_path)
    shutil.copytree(from_path, to_path)


class Export:
    def __init__(self, project_path, export_path):
        self.project_path = project_path
        self.export_path = export_path

        shutil.rmtree(self.export_path, ignore_errors=True)

        # optional: move to subdirectory to enable relative imports
        self.subdirectory = ""
        self.export_path = os.path.join(self.export_path, self.subdirectory)
        os.makedirs(self.export_path)

        self._exported = []

    @property
    def export_root(self):
        if self.subdirectory == "":
            return self.export_path
        return self.export_path[: -len(self.subdirectory)]

    def write(self, filepath, data=None, meta=False):
        target_path = self.export_path if not meta else self.export_root
        target = os.path.join(target_path, filepath)
        if data is None:
            open(target, "w").close()
            return True

        if filepath.endswith(".json"):
            with open(target, "w") as f:
                f.write(
                    json.dumps(data, indent=4, sort_keys=True, default=lambda x: str(x))
                )
            return True

        if filepath.endswith(".py"):
            with open(target, "w") as f:
                f.write(data)
            return True

    def _resolve(self, source):
        if not isinstance(source, str):
            return os.path.relpath(inspect.getfile(source), self.project_path)

        return source

    def _imports(self, path):
        Import = namedtuple("Import", ["module", "name", "alias"])

        with open(path) as fh:
            root = ast.parse(fh.read(), path)

        for node in ast.iter_child_nodes(root):
            if isinstance(node, ast.Import):
                module = ""
            elif isinstance(node, ast.ImportFrom):
                module = node.module
            else:
                continue

            for n in node.names:
                yield Import(module, n.name, n.asname)

    def _find_import(self, module, top_level=""):
        module_path = os.path.join(top_level, module.replace(".", "/"))
        module_file = module_path
        if not os.path.isdir(os.path.join(self.project_path, module_file)):
            module_file += ".py"
        if not os.path.exists(os.path.join(self.project_path, module_file)):
            return False

        return module_path, module_file

    def get_imports(self, source):
        Import = namedtuple("Import", ["module", "path", "file"])
        source_filepath = os.path.join(self.project_path, source)
        top_level = os.path.dirname(source)
        if os.path.isdir(source_filepath):
            top_level = os.path.dirname(top_level)
            source_filepath = os.path.join(source_filepath, "__init__.py")
        imports = []
        for i in self._imports(source_filepath):
            module = i.name if i.module == "" else i.module
            resolved = self._find_import(module, top_level)
            if resolved is False:
                # try again as absolute import
                resolved = self._find_import(module)
                if resolved is False:
                    continue
            imports.append(Import(module, resolved[0], resolved[1]))

        return imports

    def echo(self):
        for k in self._exported:
            print(k)

    def copy(self, source, destination=None):
        if source in self._exported:
            return source
        source = self._resolve(source)
        if destination is None:
            destination = source
        # remap recursive vendors
        if destination.count("vendor/") > 1:
            destination = "vendor/" + destination.rsplit("vendor/", 1)[-1]
        dest = os.path.join(self.export_path, destination)
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        src = os.path.join(self.project_path, source)
        cp = shutil.copyfile if not os.path.isdir(src) else copy_and_overwrite
        cp(os.path.join(self.project_path, source), dest)
        # create __init__.py files
        dirs = destination.split("/")
        for i in range(len(dirs)):
            init_py = os.path.join(self.export_path, "/".join(dirs[:i]), "__init__.py")
            if not os.path.isfile(init_py):
                open(init_py, "w").close()

        self._exported.append(destination)

        return source

    def get_import_statement(self, obj, alias=None):
        source = self._resolve(obj)
        if os.path.isdir(os.path.join(self.project_path, source)):
            module = source.replace("/", ".")
        else:
            module = source[:-3].replace("/", ".")
        if self.subdirectory != "":
            module = self.subdirectory + "." + module
        statement = f"from {module} import {obj.__name__}"
        if alias is not None and alias != obj.__name__:
            statement += " as " + alias

        return statement

    def module(self, source, from_module_path=False):
        source = self._resolve(source)
        if from_module_path:
            source = source.replace(".", "/")
            if os.path.isdir(os.path.join(self.project_path, source)):
                pass
            elif os.path.isfile(
                os.path.join(self.project_path, source.replace("/", ".") + ".py")
            ):
                source = source + ".py"
            else:
                return False
        self.copy(source)
        for imp in self.get_imports(source):
            self.module(imp.file)

        return source

    def machinable(self):
        with open(standalone.__file__, encoding="utf-8") as f:
            code = f.read()

        # parse import statements
        def parser(m):
            target = m.group(1)
            module, attribute = target.rsplit(".", 1)
            imp = importlib.import_module(module)
            obj = getattr(imp, attribute)

            source = inspect.getsource(obj)
            return source

        parsed = re.sub(r"#\s!include\s+(.*)\n", parser, code).replace(
            "Mixin = object", ""
        )

        with open(os.path.join(self.export_root, "machinable.py"), "w") as f:
            f.write(parsed)

    def entry_point(self, node, components):
        imports = self.get_import_statement(node["class"]) + "\n"
        for k, c in enumerate(components):
            imports += self.get_import_statement(c["class"]) + "\n"
        # language=python
        script = f"""\
import json

{imports}

with open('config.json') as f:
    config = json.load(f)

# load subcomponents
for k in range(len(config['components'])):
    config['components'][k]['class'] = locals()[config['components'][k]['class']]

{node['class'].__name__}(config['component']['args'], config['component']['flags']).dispatch(
    config['components'], config['storage']
)
"""

        self.write("run.py", script, meta=True)



import os

# This following method is modified 3rd party source code from
# https://github.com/IDSIA/sacred/blob/7897c664b1b93fa2e2b6f3af244dfee590b1342a/sacred/dependencies.py#L401.
# The copyright and license agreement can be found in the ThirdPartyNotices.txt file at the root of this repository.


def get_commit(filename, search_parent_directories=False):
    # git
    try:
        from git import Repo, InvalidGitRepositoryError

        try:
            directory = os.path.dirname(filename)
            repo = Repo(directory, search_parent_directories=search_parent_directories)
            try:
                path = repo.remote().url
            except ValueError:
                path = "git:/" + repo.working_dir
            is_dirty = repo.is_dirty()
            commit = repo.head.commit.hexsha
            return {"path": path, "commit": commit, "is_dirty": is_dirty}
        except (InvalidGitRepositoryError, ValueError):
            pass
    except ImportError:
        pass

    return {"path": None, "commit": None, "is_dirty": None}



import machinable as ml


def test_native_engine():
    t = ml.Experiment().components("thenode", "thechildren").repeat(2)
    ml.execute(t, engine=None, project="./test_project")


def test_native_engine_multiprocessing():
    t = ml.Experiment().components("thenode", "thechildren").repeat(5)
    ml.execute(t, engine="native:1", project="./test_project")



import pendulum


class StatusTrait:
    @property
    def started(self):
        """Returns the starting time"""
        return self._model.started

    @property
    def heartbeat(self):
        """Returns the last heartbeat time"""
        return self._model.heartbeat

    @property
    def finished(self):
        """Returns the finishing time"""
        return self._model.finished

    def is_finished(self):
        """True if finishing time has been written"""
        return None

    def is_alive(self):
        """True if last heartbeat occurred less than 30 seconds ago"""
        return self.heartbeat.substract(seconds=30) < pendulum.now()

    def is_active(self):
        """True if not finished and alive"""
        return (not self.is_finished()) and self.is_alive()



from .collection import Collection


class ObservationCollection(Collection):
    def as_dataframe(self):
        """Returns collection as Pandas dataframe
        """
        data = {k: [] for k in self._items[0].serialize().keys()}
        for observation in self._items:
            for k, v in observation.serialize().items():
                data[k].append(v)
        import pandas

        return pandas.DataFrame.from_dict(data)



import json

from machinable.config.mapping import config_map

from .status import StatusTrait


class Task(StatusTrait):
    def __init__(self, model):
        self._model = model

    @property
    def id(self):
        """6-digit experiment ID, e.g. #F4K3r6"""
        return self._model.task_id

    @property
    def execution_id(self):
        """Returns the execution ID"""
        return self._model.execution_id

    @property
    def name(self):
        """Returns the name of the experiment"""
        return self._model.name

    @property
    def seed(self):
        """Returns the global random seed used in the experiment"""
        return self._model.seed

    @property
    def tune(self):
        """True if experiment is a tuning experiment"""
        return self._model.tune

    @property
    def rerun(self):
        """Number of rerun"""
        return self._model.rerun

    @property
    def code_backup(self):
        """True if code backup is available"""
        return self._model.code_backup

    @property
    def code_version(self):
        """Returns information about the source code version as a dictionary

        ```
        project:
          path: VCS url
          commit: Commit hash or None
          is_dirty: Whether everything has been commited to VCS
        vendor: List of vendor project information with the same structure as above
        ```
        """
        return config_map(json.loads(self._model.code_version))

    @property
    def observations_count(self):
        """Returns the number of observations in this experiment"""
        return self._model.observations_count

    def __str__(self):
        return self.id

    def __repr__(self):
        return f"Experiment <{self.id}>"



import os

from fs.errors import FSError

from machinable.config.mapping import config_map
from machinable.utils.formatting import msg

from .base import BaseView
from .records import Records as RecordsView
from .status import StatusTrait
from .task import Task as TaskView


class Observation(StatusTrait, BaseView):
    @property
    def id(self):
        """Returns the observation ID"""
        return self._model.path

    @property
    def storage(self):
        """Returns the write location of the observation"""
        return self._model.storage().url

    @property
    def task(self):
        """The experiment of this observation"""
        return TaskView(self._model.task)

    @property
    def execution_id(self):
        """Returns the execution ID"""
        return self.task.execution_id

    def store(self, name=None):
        """Retrieves element from the write

        This is the counterpart to the ``Observer.write`` method.

        # Arguments
        name: Key or filename of the object that is to be retrieved. If None, a list of available objects is returned
        """
        if name is None:
            store = self._model.load_file("store.json", meta=True)
            try:
                files = self._model.filesystem().listdir("store")
                if not isinstance(store, dict):
                    store = {}
                store["$files"] = files
            except FSError:
                pass
            return store

        if os.path.splitext(name)[1] == "":
            try:
                return self._model.load_file("store.json", meta=True)[name]
            except TypeError:
                return FileNotFoundError

        return self._model.load_file(name)

    @property
    def flags(self):
        """Return the observation flags"""
        return self.component.flags

    @property
    def components(self):
        return self._lazyload(
            "components",
            lambda m: config_map(m.load_file("components.json", meta=True, default={})),
        )

    @property
    def config(self):
        """Returns the observation config"""
        return self.component.config

    @property
    def component(self):
        """Returns the component"""
        return self._lazyload(
            "component", lambda m: config_map(m.load_file("component.json", meta=True)),
        )

    @property
    def host(self):
        """Returns information on the host that produced this observation"""
        return self._lazyload(
            "host", lambda m: config_map(m.load_file("host.json", meta=True))
        )

    @property
    def log(self):
        """Returns the content of the log file"""
        return self._lazyload("log", lambda m: m.load_file("log.txt", meta=True))

    @property
    def output(self):
        """Returns the content of the log file"""
        return self._lazyload("output", lambda m: m.load_file("output.log", meta=True))

    @property
    def records(self):
        """Returns the record interface"""
        return self.get_records_writer("default")

    def get_records_writer(self, scope=None):
        """Returns a record writer

        # Arguments
        scope: The name of the record writer
        """
        if scope is None:
            # return list of available scopes
            try:
                scopes = self._model.filesystem().listdir("records")
                return [s[:-2] for s in scopes if s.endswith(".json")]
            except FSError:
                return []
        return self._lazyload(f"{scope}.records", lambda m: RecordsView(m, scope=scope))

    def __getattr__(self, item):
        # resolve child alias
        aliases = self.flags.get("COMPONENTS_ALIAS", {})
        try:
            return self.children[aliases[item]]
        except (TypeError, KeyError, IndexError):
            raise AttributeError(f"Observation has no attribute '{item}'")

    def __str__(self):
        return f"{self.id}"

    def __repr__(self):
        try:
            return f"Observation <{self.id}> ({self.task.id})"
        except AttributeError:
            return f"Observation <{self.id}>"



import datetime
import random

from machinable import Component


class Observations(Component):

    # used to generate store test data

    def on_execute_iteration(self, iteration):
        if iteration > 5:
            return StopIteration

        self.log.info(f"Observation node {self.config.id}, iteration {iteration}")
        self.record["number"] = random.randint(1, 100)
        self.record["constant"] = 42
        self.record["float"] = 13.1
        self.record["string"] = "test"
        self.record["none"] = None
        self.record["nan"] = float("NaN")
        self.record["custom"] = Observations({}, {})
        self.record["date"] = datetime.datetime.now()

        # switching type
        if iteration > 3:
            self.record["float"] = None
            self.record["string"] = 1.3

        if self.config.get("test") is True:
            # custom records
            self.store.get_record_writer("validation")["iteration"] = iteration
            self.store.get_record_writer("validation").save()

        self.store.write("test", 2, overwrite=True)
        if iteration == 1:
            self.store.write("test.txt", f"hello from observation {self.config.id}")
            self.store.write("data.json", {"observation_id": self.config.id})
            self.store.write("key", "value")
            self.store.write("test", 1, overwrite=True)



import os
import re
from typing import Callable

import yaml


class Loader(yaml.SafeLoader):
    def __init__(self, stream):
        if isinstance(stream, str):
            self._root = "./"
        else:
            # from filestream
            self._root = os.path.split(stream.name)[0]

        super(Loader, self).__init__(stream)

    def include(self, node):
        target = self.construct_scalar(node)

        if target.startswith("$/"):
            target = target[2:]

        filename = os.path.join(self._root, target)

        with open(filename, "r") as f:
            return yaml.load(f, Loader)


# Support $/ notation for includes
Loader.add_constructor("!include", Loader.include)
Loader.add_implicit_resolver("!include", re.compile(r"\$\/([^#^ ]*)"), first=None)

# Support scientific number formats
Loader.add_implicit_resolver(
    "tag:yaml.org,2002:float",
    re.compile(
        """^(?:
                             [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
                             |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
                             |\\.[0-9_]+(?:[eE][-+][0-9]+)?
                             |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
                             |[-+]?\\.(?:inf|Inf|INF)
                             |\\.(?:nan|NaN|NAN))$""",
        re.X,
    ),
    list("-+0123456789."),
)

sentinel = object()


def from_file(filename, default=sentinel):
    if not os.path.isfile(filename):
        if default is not sentinel:
            return default
        return None

    with open(filename, "r") as f:
        config = yaml.load(f, Loader)

    return config


def from_string(text):
    return yaml.load(text, Loader)


def from_callable(callable: Callable, default=sentinel):
    doc = callable.__doc__.split("machinable.yaml\n")
    if len(doc) == 1:
        if default is not sentinel:
            return default
        return None

    return from_string(doc[1])
